@@ -43,26 +43,6 @@ static const char *DEFAULT_CONFIG[] = {
  * Utilities.
  */
 
-/*!
- * \brief Recursively create directories.
- *
- * Similar to "mkdir -p".
- * * \retval 0 on success.
- * \retval <0 on error.
- */
-static int rmkdir(char *path, int mode)
-{
-	char *p = path;
-	while((p = strchr(p + 1, '/'))) {
-		*p = '\0';
-		mkdir(path, mode);
-		*p = '/';
-	}
-
-	// Final path
-	return mkdir(path, mode);
-}
-
 /* Prototypes for cf-parse.y */
 extern int cf_parse(void *scanner);
 extern int cf_get_lineno(void *scanner);
@@ -166,7 +146,7 @@ static void zone_free(conf_zone_t *zone)
  *
  * This function is called automatically after config update.
  *
- * \todo Selective hooks.
+ * \todo Selective hooks (issue #1583).
  */
 static void conf_update_hooks(conf_t *conf)
 {
@@ -192,19 +172,37 @@ static void conf_update_hooks(conf_t *conf)
 static int conf_process(conf_t *conf)
 {
 	// Check
-	if (!conf->storage) {
+	if (conf->storage == NULL) {
 		conf->storage = strdup("/var/lib/"PROJECT_EXEC);
+		if (conf->storage == NULL) {
+			return KNOTD_ENOMEM;
+		}
 	}
 	
 	// Normalize paths
 	conf->storage = strcpath(conf->storage);
+
+	// Storage directory exists?
 	struct stat st;
-	if (stat(conf->storage, &st) != 0) {
-		rmkdir(conf->storage, S_IRWXU);
+	if (stat(conf->storage, &st) == -1) {
+		log_server_error("Could not open storage directory '%s'\n", conf->storage);
+		// I assume that conf->* is freed elsewhere
+		return KNOTD_EINVAL;
+	}
+
+	// Storage directory is a directory?
+	if (S_ISDIR(st.st_mode) == 0) {
+		log_server_error("Configured storage '%s' not a directory\n", conf->storage);
+		return KNOTD_EINVAL;
 	}
 
 	// Create PID file
-	conf->pidfile = strcdup(conf->storage, "/" PID_FILE);
+	if (conf->pidfile == NULL) {
+		conf->pidfile = strcdup(conf->storage, "/" PID_FILE);
+		if (conf->pidfile == NULL) {
+			return KNOTD_ENOMEM;
+		}
+	}
 
 	// Postprocess zones
 	int ret = KNOTD_EOK;
@@ -236,35 +234,60 @@ static int conf_process(conf_t *conf)
 		if (zone->ixfr_fslimit == 0) {
 			zone->ixfr_fslimit = conf->ixfr_fslimit;
 		}
+		
+		// Relative zone filenames should be relative to storage
+		if (zone->file[0] != '/') {
+			size_t prefix_len = strlen(conf->storage) + 1; // + '\0'
+			size_t zp_len = strlen(zone->file) + 1;
+			char *ap = malloc(prefix_len + zp_len);
+			if (ap != NULL) {
+				memcpy(ap, conf->storage, prefix_len);
+				ap[prefix_len - 1] = '/';
+				memcpy(ap + prefix_len, zone->file, zp_len);
+				free(zone->file);
+				zone->file = ap;
+			} else {
+				ret = KNOTD_ENOMEM;
+				continue;
+			}
+		}
 
 		// Normalize zone filename
 		zone->file = strcpath(zone->file);
+		if (zone->file == NULL) {
+			zone->db = NULL;
+			ret = KNOTD_ENOMEM;
+			continue;
+		}
 
 		// Create zone db filename
 		size_t zname_len = strlen(zone->name);
 		size_t stor_len = strlen(conf->storage);
-		size_t size = stor_len + zname_len + 4; // db/,\0
+		size_t size = stor_len + zname_len + 4; // /db,\0
 		char *dest = malloc(size);
 		if (dest == NULL) {
 			zone->db = NULL; /* Not enough memory. */
 			ret = KNOTD_ENOMEM; /* Error report. */
 			continue;
 		}
+		char *dpos = dest;
 		
 		/* Since we have already allocd dest to accomodate
 		 * storage/zname length strcpy is safe. */
-		strncpy(dest, conf->storage, stor_len + 1);
-		if (conf->storage[stor_len - 1] != '/') {
-			strncat(dest, "/", 1);
+		memcpy(dpos, conf->storage, stor_len + 1);
+		dpos += stor_len;
+		if (*(dpos - 1) != '/') {
+			*(dpos++) = '/';
+			*dpos = '\0';
 		}
 
-		strncat(dest, zone->name, zname_len);
-		strncat(dest, "db", 2);
+		memcpy(dpos, zone->name, zname_len + 1);
+		memcpy(dpos + zname_len, "db", 3);
 		zone->db = dest;
 
 		// Create IXFR db filename
 		stor_len = strlen(conf->storage);
-		size = stor_len + zname_len + 9; // diff.db/,\0
+		size = stor_len + zname_len + 9; // /diff.db,\0
 		dest = malloc(size);
 		if (dest == NULL) {
 			zone->ixfr_db = NULL; /* Not enough memory. */
@@ -445,6 +468,8 @@ conf_t *conf_new(const char* path)
 	c->notify_timeout = CONFIG_NOTIFY_TIMEOUT;
 	c->dbsync_timeout = CONFIG_DBSYNC_TIMEOUT;
 	c->ixfr_fslimit = -1;
+	c->uid = -1;
+	c->gid = -1;
 
 	return c;
 }
@@ -472,10 +497,11 @@ int conf_parse(conf_t *conf)
 	int ret = conf_fparser(conf);
 
 	/* Postprocess config. */
-	conf_process(conf);
-
-	/* Update hooks. */
-	conf_update_hooks(conf);
+	if (ret == 0) {
+		ret = conf_process(conf);
+		/* Update hooks. */
+		conf_update_hooks(conf);
+	}
 
 	if (ret < 0) {
 		return KNOTD_EPARSEFAIL;
@@ -513,7 +539,7 @@ void conf_truncate(conf_t *conf, int unload_hooks)
 	// Unload hooks
 	if (unload_hooks) {
 		WALK_LIST_DELSAFE(n, nxt, conf->hooks) {
-			//! \todo call hook unload.
+			/*! \todo Call hook unload (issue #1583) */
 			free((conf_hook_t*)n);
 		}
 		conf->hooks_count = 0;
@@ -573,6 +599,10 @@ void conf_truncate(conf_t *conf, int unload_hooks)
 		free(conf->pidfile);
 		conf->pidfile = 0;
 	}
+	if (conf->nsid) {
+		free(conf->nsid);
+		conf->nsid = 0;
+	}
 }
 
 void conf_free(conf_t *conf)
@@ -624,11 +654,9 @@ int conf_open(const char* path)
 	}
 
 	/* Check if exists. */
-	FILE *fp = fopen(path, "r");
-	if (fp == 0) {
+	struct stat st;
+	if (stat(path, &st) != 0) {
 		return KNOTD_ENOENT;
-	} else {
-		fclose(fp);
 	}
 
 	/* Create new config. */
@@ -636,6 +664,11 @@ int conf_open(const char* path)
 
 	/* Parse config. */
 	int ret = conf_fparser(nconf);
+	if (ret == KNOTD_EOK) {
+		/* Postprocess config. */
+		ret = conf_process(nconf);
+	}
+	
 	if (ret != KNOTD_EOK) {
 		conf_free(nconf);
 		return ret;
@@ -654,9 +687,6 @@ int conf_open(const char* path)
 		}
 	}
 
-	/* Postprocess config. */
-	conf_process(nconf);
-
 	/* Synchronize. */
 	synchronize_rcu();
 
@@ -706,6 +736,10 @@ char* strcpath(char *path)
 	// Expand '~'
 	char* remainder = strchr(path,'~');
 	if (remainder != NULL) {
+		if (remainder[1] != '/') {
+			log_server_warning("Cannot expand non-login user home directory '%s', use full path instead", path);
+		}
+
 		// Get full path
 		char *tild_exp_unsafe = getenv("HOME");
 		if (tild_exp_unsafe == NULL) {
@@ -713,11 +747,12 @@ char* strcpath(char *path)
 		}
 		// Sanitize
 		size_t tild_len = strlen(tild_exp_unsafe);
-		char *tild_exp = malloc(tild_len);
+		char *tild_exp = malloc(tild_len + 1);
 		if (tild_exp == NULL) {
 			return NULL;
 		}
-		strncpy(tild_exp, tild_exp_unsafe, tild_len + 1);
+		// Duplicate tild_exp including terminating NULL
+		memcpy(tild_exp, tild_exp_unsafe, tild_len + 1);
 		if (tild_exp[tild_len - 1] == '/') {
 			tild_exp[--tild_len] = '\0';
 		}
@@ -734,11 +769,11 @@ char* strcpath(char *path)
 		
 		// Append remainder
 		++remainder;
-		size_t remainder_len = strlen(remainder);
-		strncat(npath, remainder, remainder_len);
+		strncat(npath, remainder, strlen(remainder));
+
+		free(tild_exp);
 		free(path);
 		path = npath;
-		free(tild_exp);
 	}
 
 	return path;
