@@ -19,15 +19,18 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <getopt.h>
-#include "common.h"
+#ifdef HAVE_CAP_NG_H
+#include <cap-ng.h>
+#endif /* HAVE_CAP_NG_H */
 
+#include "common.h"
+#include "common/evqueue.h"
 #include "knot/common.h"
 #include "knot/other/error.h"
 #include "knot/server/server.h"
 #include "knot/ctl/process.h"
 #include "knot/conf/conf.h"
 #include "knot/conf/logconf.h"
-#include "common/evqueue.h"
 #include "knot/server/zones.h"
 
 /*----------------------------------------------------------------------------*/
@@ -52,7 +55,6 @@ void interrupt_handle(int s)
 			sig_req_stop = 1;
 			sig_stopping = 1;
 		} else {
-			log_server_notice("OK! Exiting immediately.\n");
 			exit(1);
 		}
 	}
@@ -127,6 +129,7 @@ int main(int argc, char **argv)
 	sigemptyset(&emptyset.sa_mask);
 	emptyset.sa_flags = 0;
 	sigaction(SIGALRM, &emptyset, NULL); // Interrupt
+	sigaction(SIGPIPE, &emptyset, NULL); // Mask
 
 	// Setup event queue
 	evqueue_set(evqueue_new());
@@ -149,8 +152,8 @@ int main(int argc, char **argv)
 	// Initialize configuration
 	conf_read_lock();
 	conf_add_hook(conf(), CONF_LOG, log_conf_hook, 0);
-	conf_add_hook(conf(), CONF_LOG, zones_ns_conf_hook, server->nameserver);
-	conf_add_hook(conf(), CONF_LOG, server_conf_hook, server);
+	conf_add_hook(conf(), CONF_ALL, server_conf_hook, server);
+	conf_add_hook(conf(), CONF_ALL, zones_ns_conf_hook, server->nameserver);
 	conf_read_unlock();
 
 	// Find implicit configuration file
@@ -181,13 +184,56 @@ int main(int argc, char **argv)
 		free(cwbuf);
 		config_fn = abs_cfg;
 	}
+	
+	/* POSIX 1003.1e capabilities. */
+#ifdef HAVE_CAP_NG_H
+	
+	/* Drop all capabilities. */
+	if (capng_have_capability(CAPNG_EFFECTIVE, CAP_SETPCAP)) {
+		capng_clear(CAPNG_SELECT_BOTH);
+	
+	
+		/* Retain ability to set capabilities and FS access. */
+		capng_type_t tp = CAPNG_EFFECTIVE|CAPNG_PERMITTED;
+		capng_update(CAPNG_ADD, tp, CAP_SETPCAP);
+		capng_update(CAPNG_ADD, tp, CAP_DAC_OVERRIDE);
+		capng_update(CAPNG_ADD, tp, CAP_CHOWN); /* Storage ownership. */
+		
+		/* Allow binding to privileged ports.
+		 * (Not inheritable)
+		 */
+		capng_update(CAPNG_ADD, tp, CAP_NET_BIND_SERVICE);
+		
+		/* Allow setuid/setgid. */
+		capng_update(CAPNG_ADD, tp, CAP_SETUID);
+		capng_update(CAPNG_ADD, tp, CAP_SETGID);
+		
+		/* Allow priorities changing. */
+		capng_update(CAPNG_ADD, tp, CAP_SYS_NICE);
+		
+		/* Apply */
+		if (capng_apply(CAPNG_SELECT_BOTH) < 0) {
+			log_server_error("Couldn't set process capabilities - "
+			                 "%s.\n", strerror(errno));
+		}
+	} else {
+		log_server_info("User uid=%d is not allowed to set "
+		                "capabilities, skipping.\n", getuid());
+	}
+#endif /* HAVE_CAP_NG_H */
 
 	// Open configuration
-	log_server_info("Parsing configuration '%s' ...\n", config_fn);
-	if (conf_open(config_fn) != KNOTD_EOK) {
-
-		log_server_error("Failed to parse configuration file '%s'.\n",
-				 config_fn);
+	log_server_info("Reading configuration '%s' ...\n", config_fn);
+	int conf_ret = conf_open(config_fn);
+	if (conf_ret != KNOTD_EOK) {
+		if (conf_ret == KNOTD_ENOENT) {
+			log_server_error("Couldn't open configuration file "
+					 "'%s'.\n", config_fn);
+		} else {
+			log_server_error("Failed to parse configuration '%s'.\n",
+				config_fn);
+		}
+		server_wait(server);
 		server_destroy(&server);
 		free(config_fn);
 		return 1;
@@ -203,7 +249,7 @@ int main(int argc, char **argv)
 	// Run server
 	int res = 0;
 	log_server_info("Starting server...\n");
-	if ((res = server_start(server)) == KNOTD_EOK) {
+	if ((server_start(server)) == KNOTD_EOK) {
 
 		// Save PID
 		int has_pid = 1;
@@ -218,7 +264,9 @@ int main(int argc, char **argv)
 		if (daemonize) {
 			log_server_info("Server started as a daemon, "
 					"PID = %ld\n", (long)getpid());
-			res = chdir("/");
+			if (chdir("/") != 0) {
+				res = 1;
+			}
 		} else {
 			log_server_info("Server started in foreground, "
 					"PID = %ld\n", (long)getpid());
@@ -241,6 +289,7 @@ int main(int argc, char **argv)
 		sigaction(SIGINT,  &sa, NULL);
 		sigaction(SIGTERM, &sa, NULL);
 		sigaction(SIGHUP,  &sa, NULL);
+		sigaction(SIGPIPE, &sa, NULL);
 		sa.sa_flags = 0;
 		pthread_sigmask(SIG_BLOCK, &sa.sa_mask, NULL);
 
@@ -251,10 +300,6 @@ int main(int argc, char **argv)
 			pthread_sigmask(SIG_BLOCK, &sa.sa_mask, NULL);
 
 			/* Interrupts. */
-			/*! \todo More robust way to exit evloop.
-			 *        Event loop should exit with a special
-			 *        event.
-			 */
 			if (sig_req_stop) {
 				sig_req_stop = 0;
 				server_stop(server);
@@ -296,9 +341,10 @@ int main(int argc, char **argv)
 		}
 		pthread_sigmask(SIG_UNBLOCK, &sa.sa_mask, NULL);
 
-		if ((res = server_wait(server)) != KNOTD_EOK) {
+		if ((server_wait(server)) != KNOTD_EOK) {
 			log_server_error("An error occured while "
 					 "waiting for server to finish.\n");
+			res = 1;
 		} else {
 			log_server_info("Server finished.\n");
 		}
@@ -306,6 +352,7 @@ int main(int argc, char **argv)
 	} else {
 		log_server_fatal("An error occured while "
 				 "starting the server.\n");
+		res = 1;
 	}
 
 	// Stop server and close log
