@@ -24,17 +24,20 @@
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
+#ifdef HAVE_CAP_NG_H
+#include <cap-ng.h>
+#endif /* HAVE_CAP_NG_H */
 
 #include "common/sockaddr.h"
 #include "common/skip-list.h"
 #include "common/fdset.h"
-#include "common/prng.h"
 #include "knot/common.h"
 #include "knot/server/tcp-handler.h"
 #include "knot/server/xfr-handler.h"
 #include "knot/server/zones.h"
 #include "libknot/nameserver/name-server.h"
 #include "knot/other/error.h"
+#include "libknot/util/error.h"
 #include "libknot/util/wire.h"
 
 /* Defines */
@@ -63,7 +66,63 @@ static inline int tcp_throttle() {
 static int xfr_send_cb(int session, sockaddr_t *addr, uint8_t *msg, size_t msglen)
 {
 	UNUSED(addr);
-	return tcp_send(session, msg, msglen);
+	int ret = tcp_send(session, msg, msglen);
+	if (ret < 0) {
+		return KNOT_ECONN;
+	}
+	
+	return ret;
+}
+
+/*! \brief Send reply. */
+static int tcp_reply(int fd, uint8_t *qbuf, size_t resp_len)
+{
+	dbg_net("tcp: got answer of size %zd.\n",
+		resp_len);
+
+	int res = 0;
+	if (resp_len > 0) {
+		res = tcp_send(fd, qbuf, resp_len);
+	}
+
+	/* Check result. */
+	if (res < 0 || (size_t)res != resp_len) {
+		dbg_net("tcp: %s: failed: %d - %d.\n",
+			  "socket_send()",
+			  res, errno);
+	}
+	
+	return res;
+}
+
+/*! \brief Sweep TCP connection. */
+static void tcp_sweep(fdset_t *set, int fd, void* data)
+{
+	UNUSED(data);
+	
+	char r_addr[SOCKADDR_STRLEN] = { '\0' };
+	int r_port = 0;
+	struct sockaddr_storage addr;
+	socklen_t len = sizeof(addr);
+	getpeername(fd, (struct sockaddr*)&addr, &len);
+
+	/* Translate */
+	if (addr.ss_family == AF_INET) {
+	    struct sockaddr_in *s = (struct sockaddr_in *)&addr;
+	    r_port = ntohs(s->sin_port);
+	    inet_ntop(AF_INET, &s->sin_addr, r_addr, sizeof(r_addr));
+	} else {
+#ifndef DISABLE_IPV6
+	    struct sockaddr_in6 *s = (struct sockaddr_in6 *)&addr;
+	    r_port = ntohs(s->sin6_port);
+	    inet_ntop(AF_INET6, &s->sin6_addr, r_addr, sizeof(r_addr));
+#endif
+	}
+	
+	log_server_notice("Connection with %s:%d was terminated due to "
+	                  "inactivity.\n", r_addr, r_port);
+	fdset_remove(set, fd);
+	close(fd);
 }
 
 /*!
@@ -74,18 +133,17 @@ static int xfr_send_cb(int session, sockaddr_t *addr, uint8_t *msg, size_t msgle
  * \param w Associated I/O event.
  * \param revents Returned events.
  */
-static void tcp_handle(tcp_worker_t *w, int fd, uint8_t *qbuf, size_t qbuf_maxlen)
+static int tcp_handle(tcp_worker_t *w, int fd, uint8_t *qbuf, size_t qbuf_maxlen)
 {
 	if (fd < 0 || !w || !w->ioh) {
 		dbg_net("tcp: tcp_handle(%p, %d) - invalid parameters\n", w, fd);
-		return;
+		return KNOTD_EINVAL;
 	}
 	
 	dbg_net("tcp: handling TCP event on fd=%d in thread %p.\n",
 	        fd, (void*)pthread_self());
 
 	knot_nameserver_t *ns = w->ioh->server->nameserver;
-	xfrhandler_t *xfr_h = w->ioh->server->xfr_h;
 
 	/* Check address type. */
 	sockaddr_t addr;
@@ -93,7 +151,7 @@ static void tcp_handle(tcp_worker_t *w, int fd, uint8_t *qbuf, size_t qbuf_maxle
 		log_server_error("Socket type %d is not supported, "
 				 "IPv6 support is probably disabled.\n",
 				 w->ioh->type);
-		return;
+		return KNOTD_ECONNREFUSED;
 	}
 
 	/* Receive data. */
@@ -102,67 +160,67 @@ static void tcp_handle(tcp_worker_t *w, int fd, uint8_t *qbuf, size_t qbuf_maxle
 		dbg_net("tcp: client on fd=%d disconnected\n", fd);
 		fdset_remove(w->fdset, fd);
 		close(fd);
-		return;
+		if (n == KNOTD_EAGAIN) {
+			char r_addr[SOCKADDR_STRLEN];
+			sockaddr_tostr(&addr, r_addr, sizeof(r_addr));
+			int r_port = sockaddr_portnum(&addr);
+			log_server_warning("Couldn't receive query from %s:%d "
+			                  "within the time limit %ds.\n",
+			                  r_addr, r_port, TCP_ACTIVITY_WD);
+		}
+		return KNOTD_ECONNREFUSED;
 	}
 
 	/* Parse query. */
 //	knot_response_t *resp = knot_response_new(qbuf_maxlen);
 	size_t resp_len = qbuf_maxlen; // 64K
-
-	/* Parse query. */
 	knot_packet_type_t qtype = KNOT_QUERY_NORMAL;
-
 	knot_packet_t *packet =
 		knot_packet_new(KNOT_PACKET_PREALLOC_QUERY);
 	if (packet == NULL) {
 		uint16_t pkt_id = knot_wire_get_id(qbuf);
 		knot_ns_error_response(ns, pkt_id, KNOT_RCODE_SERVFAIL,
-				  qbuf, &resp_len);
-		return;
+				       qbuf, &resp_len);
+		tcp_reply(fd, qbuf, resp_len);
+		return KNOTD_EOK;
 	}
 
-	int res = knot_ns_parse_packet(qbuf, n, packet, &qtype);
-	if (unlikely(res != KNOTD_EOK)) {
-
-		/* Send error response on dnslib RCODE. */
-		if (res > 0) {
+	int parse_res = knot_ns_parse_packet(qbuf, n, packet, &qtype);
+	if (unlikely(parse_res != KNOT_EOK)) {
+		if (parse_res > 0) { /* Returned RCODE */
 			uint16_t pkt_id = knot_wire_get_id(qbuf);
-			knot_ns_error_response(ns, pkt_id, res,
-					  qbuf, &resp_len);
+			knot_ns_error_response(ns, pkt_id, parse_res,
+					       qbuf, &resp_len);
+			tcp_reply(fd, qbuf, resp_len);
 		}
-
-//		knot_response_free(&resp);
 		knot_packet_free(&packet);
-		return;
+		return KNOTD_EOK;
 	}
 
 	/* Handle query. */
+	int xfrt = -1;
 	knot_ns_xfr_t xfr;
-	res = KNOTD_ERROR;
+	int res = KNOTD_ERROR;
 	switch(qtype) {
 
 	/* Query types. */
 	case KNOT_QUERY_NORMAL:
 		//res = knot_ns_answer_normal(ns, packet, qbuf, &resp_len);
-		res = zones_normal_query_answer(ns, packet, &addr, qbuf, &resp_len);
+		if (zones_normal_query_answer(ns, packet, &addr,
+		                              qbuf, &resp_len) == KNOT_EOK) {
+			res = KNOTD_EOK;
+		}
 		break;
-	case KNOT_QUERY_IXFR:
-		res = xfr_request_init(&xfr, XFR_TYPE_IOUT, XFR_FLAG_TCP, packet);
-		if (res != KNOTD_EOK) {
-			knot_ns_error_response(ns, knot_packet_id(packet),
-			                       KNOT_RCODE_SERVFAIL, qbuf,
-			                       &resp_len);
-			res = KNOTD_EOK;
-			break;
-		}
-		xfr.send = xfr_send_cb;
-		xfr.session = fd;
-		memcpy(&xfr.addr, &addr, sizeof(sockaddr_t));
-		xfr_request(xfr_h, &xfr);
-		dbg_net("tcp: enqueued IXFR query on fd=%d\n", fd);
-		return;
 	case KNOT_QUERY_AXFR:
-		res = xfr_request_init(&xfr, XFR_TYPE_AOUT, XFR_FLAG_TCP, packet);
+	case KNOT_QUERY_IXFR:
+		if (qtype == KNOT_QUERY_IXFR) {
+			xfrt = XFR_TYPE_IOUT;
+		} else {
+			xfrt = XFR_TYPE_AOUT;
+		}
+		
+		/* Prepare context. */
+		res = xfr_request_init(&xfr, xfrt, XFR_FLAG_TCP, packet);
 		if (res != KNOTD_EOK) {
 			knot_ns_error_response(ns, knot_packet_id(packet),
 			                       KNOT_RCODE_SERVFAIL, qbuf,
@@ -172,12 +230,13 @@ static void tcp_handle(tcp_worker_t *w, int fd, uint8_t *qbuf, size_t qbuf_maxle
 		}
 		xfr.send = xfr_send_cb;
 		xfr.session = fd;
+		xfr.wire = qbuf;
+		xfr.wire_size = qbuf_maxlen;
 		memcpy(&xfr.addr, &addr, sizeof(sockaddr_t));
-		xfr_request(xfr_h, &xfr);
-		dbg_net("tcp: enqueued AXFR query on fd=%d\n", fd);
-		return;
 		
-	/*! \todo Implement query notify/update. */
+		/* Answer. */
+		return xfr_answer(ns, &xfr);
+		
 	case KNOT_QUERY_UPDATE:
 		knot_ns_error_response(ns, knot_packet_id(packet),
 				       KNOT_RCODE_NOTIMPL, qbuf,
@@ -185,8 +244,12 @@ static void tcp_handle(tcp_worker_t *w, int fd, uint8_t *qbuf, size_t qbuf_maxle
 		res = KNOTD_EOK;
 		break;
 		
+	case KNOT_QUERY_NOTIFY:
+		res = notify_process_request(ns, packet, &addr,
+					     qbuf, &resp_len);
+		break;
+		
 	/* Unhandled opcodes. */
-	case KNOT_QUERY_NOTIFY:    /*!< Only in UDP. */
 	case KNOT_RESPONSE_NOTIFY: /*!< Only in UDP. */
 	case KNOT_RESPONSE_NORMAL: /*!< TCP handler doesn't send queries. */
 	case KNOT_RESPONSE_AXFR:   /*!< Processed in XFR handler. */
@@ -210,25 +273,13 @@ static void tcp_handle(tcp_worker_t *w, int fd, uint8_t *qbuf, size_t qbuf_maxle
 
 	/* Send answer. */
 	if (res == KNOTD_EOK) {
-		
-		dbg_net("tcp: got answer of size %zd.\n",
-		        resp_len);
-		
-		assert(resp_len > 0);
-		res = tcp_send(fd, qbuf, resp_len);
-
-		/* Check result. */
-		if (res != (int)resp_len) {
-			dbg_net("tcp: %s: failed: %d - %d.\n",
-				  "socket_send()",
-				  res, errno);
-		}
+		tcp_reply(fd, qbuf, resp_len);
 	} else {
 		dbg_net("tcp: failed to respond to query type=%d on fd=%d - %s\n",
 		        qtype, fd, knotd_strerror(res));;
 	}
 
-	return;
+	return res;
 }
 
 static int tcp_accept(int fd)
@@ -239,6 +290,7 @@ static int tcp_accept(int fd)
 	/* Evaluate connection. */
 	if (incoming < 0) {
 		int en = errno;
+		/*! \todo Better solution so it doesn't block current connections (issue #1542). */
 		if (en != EINTR) {
 			log_server_error("Cannot accept connection "
 					 "(%d).\n", errno);
@@ -256,6 +308,13 @@ static int tcp_accept(int fd)
 		}
 	} else {
 		dbg_net("tcp: accepted connection fd=%d\n", incoming);
+		/* Set recv() timeout. */
+#ifdef SO_RCVTIMEO
+		struct timeval tv;
+		tv.tv_sec = TCP_ACTIVITY_WD;
+		tv.tv_usec = 0;
+		setsockopt(incoming, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
+#endif
 	}
 
 	return incoming;
@@ -347,7 +406,11 @@ int tcp_recv(int fd, uint8_t *buf, size_t len, sockaddr_t *addr)
 	unsigned short pktsize = 0;
 	int n = recv(fd, &pktsize, sizeof(unsigned short), MSG_WAITALL);
 	if (n < 0) {
-		return KNOTD_ERROR;
+		if (errno == EAGAIN) {
+			return KNOTD_EAGAIN;
+		} else {
+			return KNOTD_ERROR;
+		}
 	}
 
 	pktsize = ntohs(pktsize);
@@ -364,16 +427,22 @@ int tcp_recv(int fd, uint8_t *buf, size_t len, sockaddr_t *addr)
 	if (len < pktsize) {
 		return KNOTD_ENOMEM;
 	}
-
-	/* Receive payload. */
-	n = recv(fd, buf, pktsize, MSG_WAITALL);
-
+	
 	/* Get peer name. */
 	if (addr) {
 		socklen_t alen = addr->len;
 		getpeername(fd, addr->ptr, &alen);
 	}
 
+	/* Receive payload. */
+	n = recv(fd, buf, pktsize, MSG_WAITALL);
+	if (n < 0) {
+		if (errno == EAGAIN) {
+			return KNOTD_EAGAIN;
+		} else {
+			return KNOTD_ERROR;
+		}
+	}
 	dbg_net("tcp: received packet size=%d on fd=%d\n",
 		  n, fd);
 
@@ -438,6 +507,19 @@ int tcp_loop_worker(dthread_t *thread)
 		dbg_net("tcp: failed to allocate buffers for TCP worker\n");
 		return KNOTD_EINVAL;
 	}
+	
+	/* Drop all capabilities on workers. */
+#ifdef HAVE_CAP_NG_H
+	if (capng_have_capability(CAPNG_EFFECTIVE, CAP_SETPCAP)) {
+		capng_clear(CAPNG_SELECT_BOTH);
+		capng_apply(CAPNG_SELECT_BOTH);
+	}
+#endif /* HAVE_CAP_NG_H */
+	
+	/* Next sweep time. */
+	timev_t next_sweep;
+	time_now(&next_sweep);
+	next_sweep.tv_sec += TCP_SWEEP_INTERVAL;
 
 	/* Accept clients. */
 	dbg_net_verb("tcp: worker %p started\n", w);
@@ -449,8 +531,8 @@ int tcp_loop_worker(dthread_t *thread)
 		}
 
 		/* Wait for events. */
-		int nfds = fdset_wait(w->fdset);
-		if (nfds <= 0) {
+		int nfds = fdset_wait(w->fdset, (TCP_SWEEP_INTERVAL * 1000)/2);
+		if (nfds < 0) {
 			continue;
 		}
 
@@ -459,7 +541,7 @@ int tcp_loop_worker(dthread_t *thread)
 		             w, nfds);
 		fdset_it_t it;
 		fdset_begin(w->fdset, &it);
-		while(1) {
+		while(nfds > 0) {
 			
 			/* Handle incoming clients. */
 			if (it.fd == w->pipe[0]) {
@@ -472,9 +554,22 @@ int tcp_loop_worker(dthread_t *thread)
 				             "client %d\n",
 				             w, client);
 				fdset_add(w->fdset, client, OS_EV_READ);
+				fdset_set_watchdog(w->fdset, client,
+				                   TCP_HANDSHAKE_WD);
+				dbg_net("tcp: watchdog for fd=%d set to %ds\n",
+				        client, TCP_HANDSHAKE_WD);
 			} else {
 				/* Handle other events. */
-				tcp_handle(w, it.fd, qbuf, TCP_BUFFER_SIZE);
+				int ret = tcp_handle(w, it.fd, qbuf,
+				                     TCP_BUFFER_SIZE);
+				if (ret == KNOTD_EOK) {
+					fdset_set_watchdog(w->fdset, it.fd,
+					                   TCP_ACTIVITY_WD);
+					dbg_net("tcp: watchdog for fd=%d "
+					        "set to %ds\n",
+					        it.fd, TCP_ACTIVITY_WD);
+				}
+				
 			}
 			
 			/* Check if next exists. */
@@ -483,6 +578,15 @@ int tcp_loop_worker(dthread_t *thread)
 			}
 		}
 		
+		/* Sweep inactive. */
+		timev_t now;
+		if (time_now(&now) == 0) {
+			if (now.tv_sec >= next_sweep.tv_sec) {
+				fdset_sweep(w->fdset, &tcp_sweep, NULL);
+				memcpy(&next_sweep, &now, sizeof(next_sweep));
+				next_sweep.tv_sec += TCP_SWEEP_INTERVAL;
+			}
+		}
 	}
 
 	/* Stop whole unit. */
