@@ -42,7 +42,7 @@
 #include "libknot/updates/xfr-in.h"
 
 /* Constants. */
-#define SOA_QRY_TIMEOUT 10000 /*!< SOA query timeout (ms). */
+#define ZONES_JITTER_PCT    10 /*!< +-N% jitter to timers. */
 #define IXFR_DBSYNC_TIMEOUT (60*1000) /*!< Database sync timeout = 60s. */
 #define AXFR_BOOTSTRAP_RETRY (60*1000) /*!< Interval between AXFR BS retries. */
 
@@ -69,17 +69,22 @@ typedef struct zonedata_t
 	struct {
 		acl_t         *acl;      /*!< ACL for xfr-in.*/
 		sockaddr_t     master;   /*!< Master server for xfr-in.*/
+		sockaddr_t     via;      /*!< Master server transit interface.*/
 		knot_key_t    tsig_key;  /*!< Master TSIG key. */
 		struct event_t *timer;   /*!< Timer for REFRESH/RETRY. */
 		struct event_t *expire;  /*!< Timer for REFRESH. */
-		int next_id;             /*!< ID of the next awaited SOA resp.*/
 		pthread_mutex_t lock;    /*!< Pending XFR/IN lock. */
 		void           *wrkr;    /*!< Pending XFR/IN worker. */
+		int next_id;             /*!< ID of the next awaited SOA resp.*/
 		uint32_t bootstrap_retry;/*!< AXFR/IN bootstrap retry. */
+		unsigned       scheduled;/*!< Scheduled operations. */ 
 	} xfr_in;
 
 	/*! \brief List of pending NOTIFY events. */
 	list notify_pending;
+	
+	/*! \brief List of fds with pending SOA queries. */
+	int soa_pending;
 
 	/*! \brief Zone IXFR history. */
 	journal_t *ixfr_db;
@@ -121,13 +126,14 @@ int zones_update_db_from_config(const conf_t *conf, knot_nameserver_t *ns,
  * \note Current implementation rewrites the zone file.
  *
  * \param zone Evaluated zone.
+ * \param journal Journal to sync.
  *
  * \retval KNOTD_EOK if successful.
  * \retval KNOTD_ERANGE if zonefile is in sync with journal.
  * \retval KNOTD_EINVAL on invalid parameter.
  * \retval KNOTD_ERROR on unspecified error during processing.
  */
-int zones_zonefile_sync(knot_zone_t *zone);
+int zones_zonefile_sync(knot_zone_t *zone, journal_t *journal);
 
 /*!
  * \todo Document me.
@@ -209,6 +215,8 @@ int zones_ns_conf_hook(const struct conf_t *conf, void *data);
  */
 int zones_store_changesets(knot_ns_xfr_t *xfr);
 
+int zones_changesets_to_binary(knot_changesets_t *chgsets);
+
 /*!
  * \brief Load changesets from journal.
  *
@@ -232,20 +240,6 @@ int zones_store_changesets(knot_ns_xfr_t *xfr);
 int zones_xfr_load_changesets(knot_ns_xfr_t *xfr, uint32_t serial_from,
                               uint32_t serial_to);
 
-/*!
- * \brief Apply changesets to zone.
- *
- * Applies a list of XFR-style changesets to the given zone. Also checks if the
- * changesets are applicable (i.e. zone is right and has the right serial).
- *
- * \param zone Zone to which the changesets should be applied.
- * \param chsets Changesets to be applied to the zone.
- *
- * \retval KNOTD_EOK
- * \retval KNOTD_EINVAL
- */
-int zones_apply_changesets(knot_ns_xfr_t *xfr);
-
 /*!
  * \brief Update zone timers.
  *
