@@ -51,16 +51,15 @@
  * or raw data stored like this: data_len [data]
  */
 
-static inline int fwrite_to_file_crc(const void *src,
-                                     size_t size, size_t n, FILE *f,
+static inline int write_to_file_crc(const void *src,
+                                     size_t size, size_t n, int fd,
                                      crc_t *crc)
 {
-	size_t rc = fwrite(src, size, n, f);
-	if (rc != n) {
+	ssize_t rc = write(fd, src, size * n);
+	if (rc != size * n) {
 		fprintf(stderr, "fwrite: invalid write %zu (expected %zu)\n", rc,
 			n);
 	}
-	/* \todo this seems to be wrong, if you fwrite less than n items, you probably should not continue */
 
 	if (size * n > 0) {
 		*crc =
@@ -68,13 +67,16 @@ static inline int fwrite_to_file_crc(const void *src,
 		                   size * n);
 	}
 
-	/* \todo the rc return is certainly wrong as it is used in the caller function */
-//	return rc == n;
-	return (int)rc;
+	/* 
+	 * It was meant differtely, caller function does not 
+	 * care how many bytes had been written, it just cares about
+	 * success/fail (not that it is checked anyway) (#1684).
+ 	 */
+	return rc == n;
 
 }
 
-static inline int fwrite_to_stream(const void *src,
+static inline int write_to_stream(const void *src,
                                    size_t size, size_t n,
                                    uint8_t **stream,
                                    size_t *stream_size)
@@ -97,17 +99,17 @@ static inline int fwrite_to_stream(const void *src,
 	return KNOT_EOK;
 }
 
-static int fwrite_wrapper(const void *src,
-                          size_t size, size_t n, FILE *fp,
+static int write_wrapper(const void *src,
+                          size_t size, size_t n, int fd,
                           uint8_t **stream, size_t *stream_size, crc_t *crc)
 {
-	if (fp == NULL) {
+	if (fd < 0) {
 		assert(stream && stream_size);
 		assert(crc == NULL);
-		return fwrite_to_stream(src, size, n, stream, stream_size);
+		return write_to_stream(src, size, n, stream, stream_size);
 	} else {
 		assert(stream == NULL && stream_size == NULL);
-		return fwrite_to_file_crc(src, size, n, fp, crc);
+		return write_to_file_crc(src, size, n, fd, crc);
 	}
 }
 
@@ -117,17 +119,17 @@ static int fwrite_wrapper(const void *src,
  * \param dname Dname whose labels are to be dumped.
  * \param f Output file.
  */
-static void knot_labels_dump_binary(const knot_dname_t *dname, FILE *f,
+static void knot_labels_dump_binary(const knot_dname_t *dname, int fd,
                                     uint8_t **stream, size_t *stream_size,
                                     crc_t *crc)
 {
 	dbg_zdump("label count: %d\n", dname->label_count);
 	uint16_t label_count = dname->label_count;
-	/* \todo check the return value */
-	fwrite_wrapper(&label_count, sizeof(label_count), 1, f, stream,
+	/*!< \todo #1684 check the return value */
+	write_wrapper(&label_count, sizeof(label_count), 1, fd, stream,
 	               stream_size, crc);
-	/* \todo check the return value */
-	fwrite_wrapper(dname->labels, sizeof(uint8_t), dname->label_count, f,
+	/*!< \todo #1684 check the return value */
+	write_wrapper(dname->labels, sizeof(uint8_t), dname->label_count, fd,
 	               stream, stream_size, crc);
 }
 
@@ -137,31 +139,31 @@ static void knot_labels_dump_binary(const knot_dname_t *dname, FILE *f,
  * \param dname Dname to be dumped.
  * \param f Output file.
  */
-static void knot_dname_dump_binary(const knot_dname_t *dname, FILE *f,
+static void knot_dname_dump_binary(const knot_dname_t *dname, int fd,
                                    uint8_t **stream, size_t *stream_size,
                                    crc_t *crc)
 {
 	uint32_t dname_size = dname->size;
-	/* \todo check the return value */
-	fwrite_wrapper(&dname_size, sizeof(dname_size), 1, f, stream,
+	/*!< \todo #1684 check the return value */
+	write_wrapper(&dname_size, sizeof(dname_size), 1, fd, stream,
 	               stream_size, crc);
-	/* \todo check the return value */
-	fwrite_wrapper(dname->name, sizeof(uint8_t), dname->size, f,
+	/*!< \todo #1684 check the return value */
+	write_wrapper(dname->name, sizeof(uint8_t), dname->size, fd,
 	               stream, stream_size, crc);
 	dbg_zdump("dname size: %d\n", dname->size);
-	knot_labels_dump_binary(dname, f, stream, stream_size, crc);
+	knot_labels_dump_binary(dname, fd, stream, stream_size, crc);
 }
 
-/*!< \todo some global variable indicating error! */
-static void dump_dname_with_id(const knot_dname_t *dname, FILE *f,
+/*!< \todo #1684 some global variable indicating error! */
+static void dump_dname_with_id(const knot_dname_t *dname, int fd,
                                uint8_t **stream, size_t *stream_size,
                                crc_t *crc)
 {
 	uint32_t id = dname->id;
-	/* \todo check the return value */
-	fwrite_wrapper(&id, sizeof(id), 1, f, stream, stream_size, crc);
-	knot_dname_dump_binary(dname, f, stream, stream_size, crc);
-/*	if (!fwrite_wrapper_safe(&dname->id, sizeof(dname->id), 1, f)) {
+	/*!< \todo #1684 check the return value */
+	write_wrapper(&id, sizeof(id), 1, fd, stream, stream_size, crc);
+	knot_dname_dump_binary(dname, fd, stream, stream_size, crc);
+/*	if (!write_wrapper_safe(&dname->id, sizeof(dname->id), 1, f)) {
 		return KNOT_ERROR;
 	} */
 }
@@ -174,11 +176,10 @@ static void dump_dname_with_id(const knot_dname_t *dname, FILE *f,
  * \param data Arguments to be propagated.
  */
 static void knot_rdata_dump_binary(knot_rdata_t *rdata,
-				   uint32_t type, void *data, int use_ids,
+				   uint32_t type, int fd, int use_ids,
                                    uint8_t **stream, size_t *stream_size,
                                    crc_t *crc)
 {
-	FILE *f = (FILE *)((arg_t *)data)->arg1;
 	knot_rrtype_descriptor_t *desc =
 		knot_rrtype_descriptor_by_type(type);
 	assert(desc != NULL);
@@ -190,9 +191,9 @@ static void knot_rdata_dump_binary(knot_rdata_t *rdata,
 	}
 
 	/* Write rdata count. */
-	/* \todo check the return value */
-	fwrite_wrapper(&(rdata->count),
-	               sizeof(rdata->count), 1, f, stream, stream_size, crc);
+	/*!< \todo #1684 check the return value */
+	write_wrapper(&(rdata->count),
+	               sizeof(rdata->count), 1, fd, stream, stream_size, crc);
 
 	for (int i = 0; i < rdata->count; i++) {
 		if (&(rdata->items[i]) == NULL) {
@@ -215,48 +216,46 @@ static void knot_rdata_dump_binary(knot_rdata_t *rdata,
 
 			if (use_ids) {
 				/* Write ID. */
-				dbg_zload("%s \n",
-				    knot_dname_to_str(rdata->items[i].dname));
 				assert(rdata->items[i].dname->id != 0);
 
 				uint32_t id = rdata->items[i].dname->id;
-				/* \todo check the return value */
-				fwrite_wrapper(&id,
-				       sizeof(id), 1, f, stream, stream_size,
+				/*!< \todo #1684 check the return value */
+				write_wrapper(&id,
+				       sizeof(id), 1, fd, stream, stream_size,
 				               crc);
 			} else {
 //				assert(rdata->items[i].dname->id != 0);
 				dump_dname_with_id(rdata->items[i].dname,
-				                   f, stream,
+				                   fd, stream,
 				                   stream_size, crc);
 			}
 
 			/* Write in the zone bit */
 			if (rdata->items[i].dname->node != NULL && !wildcard) {
-				/* \todo check the return value */
-				fwrite_wrapper((uint8_t *)"\1",
-				       sizeof(uint8_t), 1, f, stream,
+				/*!< \todo #1684 check the return value */
+				write_wrapper((uint8_t *)"\1",
+				       sizeof(uint8_t), 1, fd, stream,
 				               stream_size, crc);
 			} else {
-				/* \todo check the return value */
-				fwrite_wrapper((uint8_t *)"\0", sizeof(uint8_t),
-				       1, f, stream, stream_size, crc);
+				/*!< \todo #1684 check the return value */
+				write_wrapper((uint8_t *)"\0", sizeof(uint8_t),
+				       1, fd, stream, stream_size, crc);
 			}
 
 			if (use_ids && wildcard) {
-				/* \todo check the return value */
-				fwrite_wrapper((uint8_t *)"\1",
-				       sizeof(uint8_t), 1, f, stream,
+				/*!< \todo #1684 check the return value */
+				write_wrapper((uint8_t *)"\1",
+				       sizeof(uint8_t), 1, fd, stream,
 				       stream_size, crc);
 				uint32_t wildcard_id = wildcard->id;
-				/* \todo check the return value */
-				fwrite_wrapper(&wildcard_id,
-				       sizeof(wildcard_id), 1, f, stream,
+				/*!< \todo #1684 check the return value */
+				write_wrapper(&wildcard_id,
+				       sizeof(wildcard_id), 1, fd, stream,
 				               stream_size, crc);
 			} else {
-				/* \todo check the return value */
-				fwrite_wrapper((uint8_t *)"\0", sizeof(uint8_t),
-				       1, f, stream,
+				/*!< \todo #1684 check the return value */
+				write_wrapper((uint8_t *)"\0", sizeof(uint8_t),
+				       1, fd, stream,
 				       stream_size, crc);
 			}
 
@@ -264,10 +263,10 @@ static void knot_rdata_dump_binary(knot_rdata_t *rdata,
 			dbg_zdump("Writing raw data. Item nr.: %d\n",
 			                 i);
 			assert(rdata->items[i].raw_data != NULL);
-			/* \todo check the return value */
-			fwrite_wrapper(rdata->items[i].raw_data,
+			/*!< \todo #1684 check the return value */
+			write_wrapper(rdata->items[i].raw_data,
 			               sizeof(uint8_t),
-			       rdata->items[i].raw_data[0] + 2, f,
+			       rdata->items[i].raw_data[0] + 2, fd,
 			               stream, stream_size, crc);
 
 			dbg_zdump("Written %d long raw data\n",
@@ -282,22 +281,25 @@ static void knot_rdata_dump_binary(knot_rdata_t *rdata,
  * \param rrsig RRSIG to be dumped.
  * \param data Arguments to be propagated.
  */
-static void knot_rrsig_set_dump_binary(knot_rrset_t *rrsig, arg_t *data,
+static void knot_rrsig_set_dump_binary(knot_rrset_t *rrsig, int fd,
                                        int use_ids,
                                        uint8_t **stream, size_t *stream_size,
                                        crc_t *crc)
 {
-	dbg_zdump("Dumping rrset \\w owner: %s\n",
-	                   knot_dname_to_str(rrsig->owner));
+dbg_zdump_exec_detail(
+	char *name = knot_dname_to_str(knot_rrset_owner(rrsig));
+	dbg_zdump("Dumping RRSIG \\w owner: %s\n",
+	                   name);
+	free(name);
+);
 	assert(rrsig->type == KNOT_RRTYPE_RRSIG);
 	assert(rrsig->rdata);
-	FILE *f = (FILE *)((arg_t *)data)->arg1;
-	/* \todo check the return value */
-	fwrite_wrapper(&rrsig->type, sizeof(rrsig->type), 1, f,
+	/*!< \todo #1684 check the return value */
+	write_wrapper(&rrsig->type, sizeof(rrsig->type), 1, fd,
 	               stream, stream_size, crc);
-	fwrite_wrapper(&rrsig->rclass, sizeof(rrsig->rclass), 1, f,
+	write_wrapper(&rrsig->rclass, sizeof(rrsig->rclass), 1, fd,
 	               stream, stream_size, crc);
-	fwrite_wrapper(&rrsig->ttl, sizeof(rrsig->ttl), 1, f,
+	write_wrapper(&rrsig->ttl, sizeof(rrsig->ttl), 1, fd,
 	               stream, stream_size, crc);
 
 	uint32_t rdata_count = 1;
@@ -308,16 +310,16 @@ static void knot_rrsig_set_dump_binary(knot_rrset_t *rrsig, arg_t *data,
 		rdata_count++;
 	}
 
-	fwrite_wrapper(&rdata_count, sizeof(rdata_count), 1, f,
+	write_wrapper(&rdata_count, sizeof(rdata_count), 1, fd,
 	               stream, stream_size, crc);
 
 	tmp_rdata = rrsig->rdata;
 	while (tmp_rdata->next != rrsig->rdata) {
-		knot_rdata_dump_binary(tmp_rdata, KNOT_RRTYPE_RRSIG, data,
+		knot_rdata_dump_binary(tmp_rdata, KNOT_RRTYPE_RRSIG, fd,
 		                         use_ids, stream, stream_size, crc);
 		tmp_rdata = tmp_rdata->next;
 	}
-	knot_rdata_dump_binary(tmp_rdata, KNOT_RRTYPE_RRSIG, data, use_ids,
+	knot_rdata_dump_binary(tmp_rdata, KNOT_RRTYPE_RRSIG, fd, use_ids,
 	                       stream, stream_size, crc);
 }
 
@@ -327,23 +329,24 @@ static void knot_rrsig_set_dump_binary(knot_rrset_t *rrsig, arg_t *data,
  * \param rrset RRSSet to be dumped.
  * \param data Arguments to be propagated.
  */
-static void knot_rrset_dump_binary(const knot_rrset_t *rrset, void *data,
+static void knot_rrset_dump_binary(const knot_rrset_t *rrset, int fd,
                                    int use_ids,
                                    uint8_t **stream, size_t *stream_size,
                                    crc_t *crc)
 {
-	FILE *f = (FILE *)((arg_t *)data)->arg1;
+	dbg_zdump_detail("zdump: rrset_dump_binary: Dumping rrset to fd=%d\n",
+	                 fd);
 
 	if (!use_ids) {
-		dump_dname_with_id(rrset->owner, f, stream, stream_size, crc);
+		dump_dname_with_id(rrset->owner, fd, stream, stream_size, crc);
 	}
 
-	/* \todo check the return value */
-	fwrite_wrapper(&rrset->type, sizeof(rrset->type), 1, f,
+	/*!< \todo #1684 check the return value */
+	write_wrapper(&rrset->type, sizeof(rrset->type), 1, fd,
 	               stream, stream_size, crc);
-	fwrite_wrapper(&rrset->rclass, sizeof(rrset->rclass), 1, f,
+	write_wrapper(&rrset->rclass, sizeof(rrset->rclass), 1, fd,
 	               stream, stream_size, crc);
-	fwrite_wrapper(&rrset->ttl, sizeof(rrset->ttl), 1, f,
+	write_wrapper(&rrset->ttl, sizeof(rrset->ttl), 1, fd,
 	               stream, stream_size, crc);
 
 	uint32_t rdata_count = 1;
@@ -356,26 +359,30 @@ static void knot_rrset_dump_binary(const knot_rrset_t *rrset, void *data,
 		rdata_count++;
 	}
 
-	fwrite_wrapper(&rdata_count, sizeof(rdata_count), 1, f,
+	write_wrapper(&rdata_count, sizeof(rdata_count), 1, fd,
 	               stream, stream_size, crc);
-	fwrite_wrapper(&has_rrsig, sizeof(has_rrsig), 1, f,
+	write_wrapper(&has_rrsig, sizeof(has_rrsig), 1, fd,
 	               stream, stream_size, crc);
+	
+	dbg_zdump_detail("zdump: rrset_dump_binary: Static data dumped.\n");
 
 	tmp_rdata = rrset->rdata;
 
 	while (tmp_rdata->next != rrset->rdata) {
-		knot_rdata_dump_binary(tmp_rdata, rrset->type, data, use_ids,
+		knot_rdata_dump_binary(tmp_rdata, rrset->type, fd, use_ids,
 		                       stream, stream_size, crc);
 		tmp_rdata = tmp_rdata->next;
 	}
-	knot_rdata_dump_binary(tmp_rdata, rrset->type, data, use_ids,
+	knot_rdata_dump_binary(tmp_rdata, rrset->type, fd, use_ids,
 	                       stream, stream_size, crc);
+	
+	dbg_zdump_detail("zdump: rrset_dump_binary: Rdata dumped.\n");
 
 	/* This is now obsolete, although I'd rather not use recursion - that
 	 * would probably not work */
 
 	if (rrset->rrsigs != NULL) {
-		knot_rrsig_set_dump_binary(rrset->rrsigs, data, use_ids,
+		knot_rrsig_set_dump_binary(rrset->rrsigs, fd, use_ids,
 		                           stream, stream_size, crc);
 	}
 }
@@ -386,52 +393,56 @@ static void knot_rrset_dump_binary(const knot_rrset_t *rrset, void *data,
  * \param node Node to dumped.
  * \param data Arguments to be propagated.
  */
-static void knot_node_dump_binary(knot_node_t *node, void *data,
+static void knot_node_dump_binary(knot_node_t *node, int fd,
                                   uint8_t **stream, size_t *stream_size,
                                   crc_t *crc)
 {
-	arg_t *args = (arg_t *)data;
-	FILE *f = (FILE *)args->arg1;
-
-//	node_count++;
+	if (node == NULL) {
+		return;
+	}
+	
 	/* first write dname */
 	assert(node->owner != NULL);
 
 	/* Write owner ID. */
+dbg_zdump_exec_detail(
+	char *name = knot_dname_to_str(knot_node_owner(node));
 	dbg_zdump("Dumping node owned by %s\n",
-	                   knot_dname_to_str(node->owner));
+	                   name);
+	free(name);
+);
 	assert(node->owner->id != 0);
 	uint32_t owner_id = node->owner->id;
-	/* \todo check the return value */
-	fwrite_wrapper(&owner_id, sizeof(owner_id), 1, f, stream, stream_size,
+	/*!< \todo #1684 check the return value */
+	write_wrapper(&owner_id, sizeof(owner_id), 1, fd, stream, stream_size,
 	               crc);
 
-	if (knot_node_parent(node, 0) != NULL) {
+	if (knot_node_parent(node) != NULL) {
 		uint32_t parent_id = knot_dname_id(
-				knot_node_owner(knot_node_parent(node, 0)));
-		fwrite_wrapper(&parent_id, sizeof(parent_id), 1, f,
+				knot_node_owner(knot_node_parent(node)));
+		write_wrapper(&parent_id, sizeof(parent_id), 1, fd,
 		               stream, stream_size, crc);
 	} else {
 		uint32_t parent_id = 0;
-		fwrite_wrapper(&parent_id, sizeof(parent_id), 1, f,
+		write_wrapper(&parent_id, sizeof(parent_id), 1, fd,
 		               stream, stream_size, crc);
 	}
 
-	fwrite_wrapper(&(node->flags), sizeof(node->flags), 1, f,
+	write_wrapper(&(node->flags), sizeof(node->flags), 1, fd,
 	               stream, stream_size, crc);
 
 	dbg_zdump("Written flags: %u\n", node->flags);
 
-	if (knot_node_nsec3_node(node, 0) != NULL) {
+	if (knot_node_nsec3_node(node) != NULL) {
 		uint32_t nsec3_id =
-			knot_node_owner(knot_node_nsec3_node(node, 0))->id;
-		fwrite_wrapper(&nsec3_id, sizeof(nsec3_id), 1, f,
+			knot_node_owner(knot_node_nsec3_node(node))->id;
+		write_wrapper(&nsec3_id, sizeof(nsec3_id), 1, fd,
 		               stream, stream_size, crc);
 		dbg_zdump("Written nsec3 node id: %u\n",
-			 knot_node_owner(knot_node_nsec3_node(node, 0))->id);
+			 knot_node_owner(knot_node_nsec3_node(node))->id);
 	} else {
 		uint32_t nsec3_id = 0;
-		fwrite_wrapper(&nsec3_id, sizeof(nsec3_id), 1, f,
+		write_wrapper(&nsec3_id, sizeof(nsec3_id), 1, fd,
 		               stream, stream_size, crc);
 	}
 
@@ -439,49 +450,21 @@ static void knot_node_dump_binary(knot_node_t *node, void *data,
 	 * but that number is yet unknown */
 
 	uint16_t rrset_count = node->rrset_count;
-	fwrite_wrapper(&rrset_count, sizeof(rrset_count), 1, f,
+	write_wrapper(&rrset_count, sizeof(rrset_count), 1, fd,
 	               stream, stream_size, crc);
 
-	/* CLEANUP */
-//	const skip_node_t *skip_node = skip_first(node->rrsets);
-
 	const knot_rrset_t **node_rrsets = knot_node_rrsets(node);
 	for (int i = 0; i < rrset_count; i++)
 	{
-		knot_rrset_dump_binary(node_rrsets[i], data, 1,
+		knot_rrset_dump_binary(node_rrsets[i], fd, 1,
 		                       stream, stream_size, crc);
 	}
 
-	/* CLEANUP */
-//	if (skip_node == NULL) {
-//		/* we can return, count is set to 0 */
-//		return;
-//	}
-
-//	knot_rrset_t *tmp;
-
-//	do {
-//		tmp = (knot_rrset_t *)skip_node->value;
-//		knot_rrset_dump_binary(tmp, data, 1);
-//	} while ((skip_node = skip_next(skip_node)) != NULL);
-
 	free(node_rrsets);
 
-	dbg_zdump("Position after all rrsets: %ld\n", ftell(f));
-	dbg_zdump("Writing here: %ld\n", ftell(f));
-	dbg_zdump("Function ends with: %ld\n\n", ftell(f));
 }
 
-/*!
- * \brief Checks if zone uses DNSSEC and/or NSEC3
- *
- * \param zone Zone to be checked.
- *
- * \retval 0 if zone is not secured.
- * \retval 2 if zone uses NSEC3
- * \retval 1 if zone uses NSEC
- */
-static int zone_is_secure(knot_zone_contents_t *zone)
+int zone_is_secure(knot_zone_contents_t *zone)
 {
 	if (knot_node_rrset(knot_zone_contents_apex(zone),
 			      KNOT_RRTYPE_DNSKEY) == NULL) {
@@ -496,50 +479,32 @@ static int zone_is_secure(knot_zone_contents_t *zone)
 	}
 }
 
-/*!
- * \brief Safe wrapper around fwrite.
- *
- * \param dst Destination pointer.
- * \param size Size of element to be written.
- * \param n Number of elements to be written.
- * \param fp File to write to.
- *
- * \retval > 0 if succesfull.
- * \retval 0 if failed.
- */
-//static inline int fwrite_wrapper_safe(const void *src,
-//                                      size_t size, size_t n, FILE *fp)
-//{
-//	int rc = fwrite_wrapper(src, size, n, fp);
-//	if (rc != n) {
-//		fprintf(stderr, "fwrite_wrapper: invalid write %d (expected %zu)\n", rc,
-//			n);
-//	}
-
-//	return rc == n;
-//}
-
 static void dump_dname_from_tree(knot_dname_t *dname,
 				 void *data)
 {
 	arg_t *arg = (arg_t *)data;
-	FILE *f = (FILE *)arg->arg1;
+	int *fd_pointer = (int *)arg->arg1;
+	int fd = -1;
+	if (fd_pointer != NULL) {
+		fd = *fd_pointer;
+	} else {
+		dbg_zdump("zdump: dump_dname_from_tree: Bad fd.\n");
+		return;
+	}
+	
 	crc_t *crc = (crc_t*)arg->arg2;
-	dump_dname_with_id(dname, f, NULL, NULL, crc);
+	dump_dname_with_id(dname, fd, NULL, NULL, crc);
 }
 
 static int knot_dump_dname_table(const knot_dname_table_t *dname_table,
-				   FILE *f, crc_t *crc)
+				   int fd, crc_t *crc)
 {
 	arg_t arg;
-	arg.arg1 = f;
+	arg.arg1 = &fd;
 	arg.arg2 = crc;
 	/* Go through the tree and dump each dname along with its ID. */
 	knot_dname_table_tree_inorder_apply(dname_table,
 					    dump_dname_from_tree, &arg);
-	/* CLEANUP */
-//	TREE_FORWARD_APPLY(dname_table->tree, dname_table_node, avl,
-//			   dump_dname_from_tree, (void *)f);
 
 	return KNOT_EOK;
 }
@@ -559,27 +524,38 @@ static void save_node_from_tree(knot_node_t *node, void *data)
 static void dump_node_to_file(knot_node_t *node, void *data)
 {
 	arg_t *arg = (arg_t *)data;
-	knot_node_dump_binary(node, data, NULL, NULL, (crc_t *)arg->arg7);
+	int *fd_pointer = (int *)arg->arg1;
+	int fd = -1;
+	if (fd_pointer != NULL) {
+		fd = *fd_pointer;
+	}
+	
+	knot_node_dump_binary(node, fd, NULL, NULL, (crc_t *)arg->arg7);
 }
 
-int knot_zdump_binary(knot_zone_contents_t *zone, const char *filename,
-			int do_checks, const char *sfilename)
+char *knot_zdump_crc_file(const char* filename)
 {
-	/* Open .new file. */
-	char new_path[strlen(filename) + strlen(".new") + 1];
-	memcpy(new_path, filename, strlen(filename) + 1);
-	strcat(new_path, ".new");
-	mode_t mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
-	int fd = open(new_path, O_WRONLY | O_CREAT | O_TRUNC, mode);
-	if (fd == -1) {
+	char *crc_path =
+		malloc(sizeof(char) * (strlen(filename) +
+		strlen(".crc") + 1));
+	CHECK_ALLOC_LOG(crc_path, NULL);
+	memset(crc_path, 0,
+	       sizeof(char) * (strlen(filename) +
+	       strlen(".crc") + 1));
+	memcpy(crc_path, filename,
+	       sizeof(char) * strlen(filename));
+	crc_path = strncat(crc_path, ".crc", strlen(".crc"));
+	return crc_path;
+}
+
+int knot_zdump_binary(knot_zone_contents_t *zone, int fd,
+                      int do_checks, const char *sfilename,
+                      crc_t *crc)
+{
+	if (fd < 0 || sfilename == NULL) {
 		return KNOT_EBADARG;
 	}
 
-	FILE *f = fdopen(fd, "wb");
-	assert(f);
-
-	/* CLEANUP */
-//	skip_list_t *encloser_list = skip_create_list(compare_pointers);
 	arg_t arguments;
 	/* Memory to be derefenced in the save_node_from_tree function. */
 	uint32_t node_count = 0;
@@ -607,146 +583,85 @@ int knot_zdump_binary(knot_zone_contents_t *zone, const char *filename,
 		do_checks += zone_is_secure(zone);
 	}
 
-	err_handler_t *handler = NULL;
-
-	if (do_checks) {
-		handler = handler_new(1, 0, 1, 1, 1);
-		if (handler == NULL) {
-			/* disable checks and we can continue */
-			do_checks = 0;
-		} else { /* Do check for SOA right now */
-			if (knot_node_rrset(knot_zone_contents_apex(zone),
-					      KNOT_RRTYPE_SOA) == NULL) {
-				err_handler_handle_error(handler,
-							 knot_zone_contents_apex(zone),
-							 ZC_ERR_MISSING_SOA);
-			}
+	/* FIXME(OS): Really descriptive call 1,1,1,1, some #defines here? */
+	err_handler_t *handler = handler_new(1, 1, 1, 1, 1);
+	if (handler == NULL) {
+		return KNOT_ENOMEM;
+	} else { /* Do check for SOA right now */
+		if (knot_node_rrset(knot_zone_contents_apex(zone),
+				      KNOT_RRTYPE_SOA) == NULL) {
+			err_handler_handle_error(handler,
+						 knot_zone_contents_apex(zone),
+						 ZC_ERR_MISSING_SOA);
 		}
-
-		knot_node_t *last_node = NULL;
-
-		zone_do_sem_checks(zone, do_checks, handler, &last_node);
-		log_cyclic_errors_in_zone(handler, zone, last_node,
-		                          first_nsec3_node, last_nsec3_node,
-		                          do_checks);
-		err_handler_log_all(handler);
-		free(handler);
+	}
+	
+	knot_node_t *last_node = NULL;
+	int ret = zone_do_sem_checks(zone,
+	                             do_checks, handler, &last_node);
+	log_cyclic_errors_in_zone(handler, zone, last_node,
+	                          first_nsec3_node, last_nsec3_node,
+	                          do_checks);
+	err_handler_log_all(handler);
+	free(handler);
+		
+	if (ret != KNOT_EOK) {
+		fprintf(stderr, "Zone will not be dumped because of "
+		        "fatal semantic errors.\n");
+		/* If remove fails, there is nothing we can do. */
+		return KNOT_ERROR;
 	}
 
-	crc_t crc = crc_init();
+	*crc = crc_init();
 
 	/* Start writing header - magic bytes. */
 	static const uint8_t MAGIC[MAGIC_LENGTH] = MAGIC_BYTES;
-	fwrite_wrapper(&MAGIC, sizeof(uint8_t), MAGIC_LENGTH, f, NULL, NULL,
-	               &crc);
+	write_wrapper(&MAGIC, sizeof(uint8_t), MAGIC_LENGTH, fd, NULL, NULL,
+	               crc);
 
 	/* Write source file length. */
-	uint32_t sflen = 0;
-	if (sfilename) {
-		sflen = strlen(sfilename) + 1;
-	}
-	fwrite_wrapper(&sflen, sizeof(uint32_t), 1, f, NULL, NULL, &crc);
+	uint32_t sflen = strlen(sfilename) + 1;
+	write_wrapper(&sflen, sizeof(uint32_t), 1, fd, NULL, NULL, crc);
 
 	/* Write source file. */
-	fwrite_wrapper(sfilename, sflen, 1, f, NULL, NULL, &crc);
+	write_wrapper(sfilename, sflen, 1, fd, NULL, NULL, crc);
 
 	/* Notice: End of header,
 	 */
 
 	/* Start writing compiled data. */
-	fwrite_wrapper(&normal_node_count, sizeof(normal_node_count), 1, f,
-	               NULL, NULL, &crc);
-	fwrite_wrapper(&nsec3_node_count, sizeof(nsec3_node_count), 1, f,
-	               NULL, NULL, &crc);
+	write_wrapper(&normal_node_count, sizeof(normal_node_count), 1, fd,
+	               NULL, NULL, crc);
+	write_wrapper(&nsec3_node_count, sizeof(nsec3_node_count), 1, fd,
+	               NULL, NULL, crc);
 	uint32_t auth_node_count = zone->node_count;
-	fwrite_wrapper(&auth_node_count,
-	       sizeof(auth_node_count), 1, f, NULL, NULL, &crc);
+	write_wrapper(&auth_node_count,
+	       sizeof(auth_node_count), 1, fd, NULL, NULL, crc);
 
 	/* Write total number of dnames */
 	assert(zone->dname_table);
 	uint32_t total_dnames = zone->dname_table->id_counter;
-	fwrite_wrapper(&total_dnames,
-	       sizeof(total_dnames), 1, f, NULL, NULL, &crc);
+	write_wrapper(&total_dnames,
+	       sizeof(total_dnames), 1, fd, NULL, NULL, crc);
 
 	/* Write dname table. */
-	if (knot_dump_dname_table(zone->dname_table, f, &crc)
+	if (knot_dump_dname_table(zone->dname_table, fd, crc)
 	    != KNOT_EOK) {
 		return KNOT_ERROR;
 	}
-
-	arguments.arg1 = (void *)f;
+	
+	arguments.arg1 = &fd;
 	arguments.arg3 = zone;
-	arguments.arg7 = &crc;
+	arguments.arg7 = crc;
 
-	/* TODO is there a way how to stop the traversal upon error? */
+	/*!< \todo #1685 Stop traversal upon error. */
 	knot_zone_contents_tree_apply_inorder(zone, dump_node_to_file,
 				       (void *)&arguments);
 
 	knot_zone_contents_nsec3_apply_inorder(zone, dump_node_to_file,
 					(void *)&arguments);
-	fclose(f);
-
-	crc = crc_finalize(crc);
-	/* Write CRC to separate .crc file. */
-	/*!< \todo There is now function doing this. */
-	char *crc_path =
-		malloc(sizeof(char) * (strlen(filename) + strlen(".crc") + 1));
-	if (unlikely(!crc_path)) {
-		close(fd);
-		return KNOT_ENOMEM;
-	}
-	memset(crc_path, 0,
-	       sizeof(char) * (strlen(filename) + strlen(".crc") + 1));
-	memcpy(crc_path, filename, sizeof(char) * strlen(filename));
-
-	crc_path = strcat(crc_path, ".crc");
-	FILE *f_crc = fopen(crc_path, "w");
-	if (unlikely(!f_crc)) {
-		dbg_zload("knot_zload_open: failed to open '%s'\n",
-		                   crc_path);
-		close(fd);
-		free(crc_path);
-		return ENOENT;
-	}
-	free(crc_path);
-
-	fprintf(f_crc, "%lu\n", (unsigned long)crc);
-	fclose(f_crc);
-
-	close(fd);
-	mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
-	fd = open(filename, O_WRONLY | O_CREAT, mode);
-	if (fd == -1) {
-		fprintf(stderr, "%s\n", strerror(errno));
-		fprintf(stderr, "Could not open destination file! Use '%s' "
-		        "file instead.\n", new_path);
-		return KNOT_ERROR;
-	}
-
-	/* Try to obtain exclusive lock for originally given file. */
-	if (fcntl(fd, F_SETLK, knot_file_lock(F_WRLCK, SEEK_SET)) == -1) {
-		fprintf(stderr, "Could not lock destination file for write! "
-		        "Use '%s' file instead.\n", new_path);
-		close(fd);
-		return KNOT_ERROR;
-	}
-
-	/* Move .new file to original file. */
-	if (rename(new_path, filename) != 0) {
-		fprintf(stderr, "Could not move to originally given file! "
-		        "Use '%s' file instead.\n", new_path);
-		close(fd);
-		return KNOT_ERROR;
-	}
-
-	/* Release the lock. */
-	if (fcntl(fd, F_SETLK, knot_file_lock(F_UNLCK, SEEK_SET)) == -1) {
-		fprintf(stderr, "Could not unlock destination file!\n");
-		return KNOT_ERROR;
-	}
-
-	close(fd);
-
+	*crc = crc_finalize(*crc);
+	
 	return KNOT_EOK;
 }
 
@@ -755,89 +670,28 @@ int knot_zdump_rrset_serialize(const knot_rrset_t *rrset, uint8_t **stream,
 {
 	if (stream == NULL || *stream != NULL || rrset == NULL ||
 	    size == NULL) {
+		dbg_zdump("zdump: rrset_serialize: Bad arguments.\n");
 		return KNOT_EBADARG;
 	}
 
 	*size = 0;
 	arg_t arguments;
 	memset(&arguments, 0, sizeof(arg_t));
+	
+	/* This fd will signal functions to use streams. */
+	int fd = -1;
 
-	knot_rrset_dump_binary(rrset, &arguments, 0, stream, size, NULL);
+	knot_rrset_dump_binary(rrset, fd, 0, stream, size, NULL);
 
 	return KNOT_EOK;
 }
 
-static char *knot_zdump_crc_file(const char* filename)
+int knot_zdump_dump(knot_zone_contents_t *zone, int fd, const char *sfilename,
+                    crc_t *crc)
 {
-	char *crc_path =
-		malloc(sizeof(char) * (strlen(filename) +
-		strlen(".crc") + 1));
-	CHECK_ALLOC_LOG(crc_path, NULL);
-	memset(crc_path, 0,
-	       sizeof(char) * (strlen(filename) +
-	       strlen(".crc") + 1));
-	memcpy(crc_path, filename,
-	       sizeof(char) * strlen(filename));
-	crc_path = strcat(crc_path, ".crc");
-	return crc_path;
-}
-
-int knot_zdump_dump_and_swap(knot_zone_contents_t *zone,
-                             const char *temp_zonedb,
-                             const char *destination_zonedb,
-                             const char *sfilename)
-{
-	int rc = knot_zdump_binary(zone, temp_zonedb, 0, sfilename);
-
+	int rc = knot_zdump_binary(zone, fd, 0, sfilename, crc);
 	if (rc != KNOT_EOK) {
-		dbg_zdump("Failed to save the zone to binary zone db %s."
-		                 "\n", temp_zonedb);
-		return KNOT_ERROR;
-	}
-
-	/*! \todo this would also need locking as well. */
-	rc = remove(destination_zonedb);
-	if (rc == 0 || (rc != 0 && errno == ENOENT)) {
-
-		/* Delete old CRC file. */
-		char *destination_zonedb_crc =
-			knot_zdump_crc_file(destination_zonedb);
-		if (destination_zonedb_crc == NULL) {
-			return KNOT_ENOMEM;
-		}
-		remove(destination_zonedb_crc);
-
-		/* Move CRC file. */
-		char *temp_zonedb_crc =
-			knot_zdump_crc_file(temp_zonedb);
-		if (temp_zonedb_crc == NULL) {
-			return KNOT_ENOMEM;
-		}
-
-		if (rename(temp_zonedb_crc, destination_zonedb_crc) != 0) {
-			dbg_zdump("Failed to replace old zonedb CRC %s "
-					 "with new CRC zone file %s.\n",
-					 destination_zonedb_crc,
-					 temp_zonedb_crc);
-			return KNOT_ERROR;
-		}
-		free(temp_zonedb_crc);
-		free(destination_zonedb_crc);
-
-		/* Rename zonedb. */
-		if (rename(temp_zonedb, destination_zonedb) != 0) {
-			dbg_zdump("Failed to replace old zonedb %s "
-					 "with new zone file %s.\n",
-					 temp_zonedb,
-					 destination_zonedb);
-			/*! \todo with proper locking, this shouldn't happen,
-			 *        revise it later on.
-			 */
-			return KNOT_ERROR;
-		}
-	} else {
-		dbg_zdump("Failed to replace old zonedb '%s'', %s.\n",
-				 destination_zonedb, strerror(errno));
+		dbg_zdump("Failed to save the zone to binary zone db\n.");
 		return KNOT_ERROR;
 	}
 
