@@ -20,8 +20,11 @@
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
+#include <sys/types.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #include <time.h>
+#include <unistd.h>
 
 #include "common/crc.h"
 #include "libknot/common.h"
@@ -235,6 +238,9 @@ static knot_rdata_t *knot_load_rdata(uint16_t type, FILE *f,
                                          int use_ids)
 {
 	knot_rdata_t *rdata = knot_rdata_new();
+	if (rdata == NULL) {
+		return NULL;
+	}
 
 	knot_rrtype_descriptor_t *desc =
 		knot_rrtype_descriptor_by_type(type);
@@ -246,6 +252,7 @@ static knot_rdata_t *knot_load_rdata(uint16_t type, FILE *f,
 	uint32_t rdata_count = 0;
 
 	if(!fread_wrapper(&rdata_count, sizeof(rdata_count), 1, f)) {
+		knot_rdata_free(&rdata);
 		return NULL;
 	}
 
@@ -267,7 +274,9 @@ static knot_rdata_t *knot_load_rdata(uint16_t type, FILE *f,
 		desc->wireformat[i] == KNOT_RDATA_WF_UNCOMPRESSED_DNAME ||
 		desc->wireformat[i] == KNOT_RDATA_WF_LITERAL_DNAME )	{
 
-			/* TODO maybe this does not need to be stored this big*/
+			/*!< \todo #1686
+			 * Refactor these variables, some might be too big.
+			 */
 
 			uint32_t dname_id = 0;
 			uint8_t has_wildcard = 0;
@@ -310,7 +319,7 @@ static knot_rdata_t *knot_load_rdata(uint16_t type, FILE *f,
 			} else if (use_ids && !in_the_zone) { /* destroy the node */
 				if (id_array[dname_id]->node != NULL) {
 					knot_node_free(&id_array[dname_id]->
-							 node, 0, 0);
+							 node, 0);
 				}
 				/* Also sets node to NULL! */
 			}
@@ -321,8 +330,9 @@ static knot_rdata_t *knot_load_rdata(uint16_t type, FILE *f,
 				load_rdata_purge(rdata, items, i, desc, type);
 				return NULL;
 			}
-
-			items[i].raw_data =
+			
+			/*!< \todo this is not proper fix, see #1678 */
+			items[i].raw_data = (uint16_t *)
 				malloc(sizeof(uint8_t) * (raw_data_length + 2));
 			items[i].raw_data[0] = raw_data_length;
 
@@ -441,35 +451,60 @@ static knot_rrset_t *knot_load_rrset(FILE *f, knot_dname_t **id_array,
 	}
 
 	if (!fread_wrapper(&rrset_type, sizeof(rrset_type), 1, f)) {
+		if (!use_ids) {
+			knot_dname_free(&owner);
+		}
 		return NULL;
 	}
 	dbg_zload("Zone load: rrset load: type: %u\n", rrset_type);
 	if (!fread_wrapper(&rrset_class, sizeof(rrset_class), 1, f)) {
+		if (!use_ids) {
+			knot_dname_free(&owner);
+		}
 		return NULL;
 	}
 	dbg_zload("Zone load: rrset class: type: %u\n", rrset_class);
 	if (!fread_wrapper(&rrset_ttl, sizeof(rrset_ttl), 1, f)) {
+		if (!use_ids) {
+			knot_dname_free(&owner);
+		}
 		return NULL;
 	}
 	dbg_zload("Zone load: rrset ttl: type: %u\n", rrset_ttl);
 	if (!fread_wrapper(&rdata_count, sizeof(rdata_count), 1, f)) {
+		if (!use_ids) {
+			knot_dname_free(&owner);
+		}
 		return NULL;
 	}
 	dbg_zload("Zone load: rrset load: rdata count: %u\n", rdata_count);
 	if (!fread_wrapper(&rrsig_count, sizeof(rrsig_count), 1, f)) {
+		if (!use_ids) {
+			knot_dname_free(&owner);
+		}
 		return NULL;
 	}
 	dbg_zload("Zone load: rrset load: type: %u\n", rrset_type);
 
+dbg_zload_exec_detail(
+	char *name = knot_dname_to_str(owner);
 	dbg_zload("Loading RRSet owned by: %s\n",
-	          knot_dname_to_str(owner));
+	          name);
+	free(name);
+);
 
 	rrset = knot_rrset_new(owner, rrset_type, rrset_class, rrset_ttl);
+	
+	if (rrset == NULL) {
+		dbg_zload("zload: load_rrset: Could not create rrset.");
+		knot_dname_free(&owner);
+		return NULL;
+	}
 
 	if (!use_ids) {
 		/* Directly release if allocated locally. */
 		knot_dname_release(owner);
-		owner = 0;
+		owner = NULL;
 	}
 
 	dbg_zload("RRSet type: %d\n", rrset->type);
@@ -571,7 +606,7 @@ static knot_node_t *knot_load_node(FILE *f, knot_dname_t **id_array)
 
 	if (parent_id != 0) {
 		knot_node_set_parent(node, id_array[parent_id]->node);
-		assert(knot_node_parent(node, 0) != NULL);
+		assert(knot_node_parent(node) != NULL);
 	} else {
 		knot_node_set_parent(node, NULL);
 	}
@@ -580,8 +615,10 @@ static knot_node_t *knot_load_node(FILE *f, knot_dname_t **id_array)
 
 	for (int i = 0; i < rrset_count; i++) {
 		if ((tmp_rrset = knot_load_rrset(f, id_array, 1)) == NULL) {
-			knot_node_free(&node, 1, 0);
-			//TODO what else to free?
+			knot_node_free(&node, 0);
+			/*!< \todo #1686 
+			 * Refactor freeing, might not be enough.
+			 */
 			fprintf(stderr, "zone: Could not load rrset.\n");
 			return NULL;
 		}
@@ -611,7 +648,6 @@ static void find_and_set_wildcard_child(knot_zone_contents_t *zone,
 				 knot_node_t *node, int nsec3)
 {
 	knot_dname_t *chopped = knot_dname_left_chop(node->owner);
-	assert(chopped);
 	knot_node_t *wildcard_parent;
 	if (!nsec3) {
 		wildcard_parent =
@@ -665,8 +701,8 @@ static unsigned long calculate_crc(FILE *f)
 	size_t file_size = ftell(f);
 	fseek(f, 0L, SEEK_SET);
 
-	const size_t chunk_size = 1024;
-	/* read chunks of 1 kB */
+	const size_t chunk_size = 4096;
+	/* read chunks of 4 kB */
 	size_t read_bytes = 0;
 	/* Prealocate chunk */
 	unsigned char *chunk = malloc(sizeof(unsigned char) * chunk_size);
@@ -698,18 +734,29 @@ static unsigned long calculate_crc(FILE *f)
 
 int knot_zload_open(zloader_t **dst, const char *filename)
 {
-	*dst = 0;
+	char crc_buf[65];
+
 	if (!dst || !filename) {
 		return KNOT_EBADARG;
 	}
 
+	*dst = 0;
+
 	fread_wrapper = fread_safe_from_file;
 
 	/* Open file for binary read. */
 	FILE *f = fopen(filename, "rb");
 	if (unlikely(!f)) {
+		int reason = errno;
 		dbg_zload("knot_zload_open: failed to open '%s'\n",
 				   filename);
+		switch (reason) {
+		case EACCES: return KNOT_EACCES; break;
+		case ENOENT: return KNOT_ENOENT; break;
+		case ENOMEM: return KNOT_ENOMEM; break;
+		default: break;
+		}
+
 		return KNOT_EFEWDATA; // No such file or directory (POSIX.1)
 	}
 
@@ -718,38 +765,56 @@ int knot_zload_open(zloader_t **dst, const char *filename)
 
 	/* Read CRC from filename.crc file */
 	char *crc_path =
-		malloc(sizeof(char) * (strlen(filename) + strlen(".crc") + 1));
+		malloc(sizeof(char) * (strlen(filename) + 4 /* strlen(".crc") */ + 1));
 	if (unlikely(!crc_path)) {
 		fclose(f);
 		return KNOT_ENOMEM;
 	}
-	memset(crc_path, 0,
-	       sizeof(char) * (strlen(filename) + strlen(".crc") + 1));
+	memcpy(crc_path, filename, strlen(filename));
+	memcpy(crc_path + strlen(filename), ".crc", 4);
+	crc_path[strlen(filename) + 4] = '\0';
 
-	memcpy(crc_path, filename, sizeof(char) * strlen(filename));
-
-	crc_path = strcat(crc_path, ".crc");
-	FILE *f_crc = fopen(crc_path, "r");
-	if (unlikely(!f_crc)) {
+	int f_crc = open(crc_path, O_RDONLY);
+	if (f_crc == -1) {
+		/* FIXME: Print strerror_r(errno) in dbg message */
 		dbg_zload("knot_zload_open: failed to open '%s'\n",
-		                   crc_path);
-		fclose(f);
+			  crc_path);
 		free(crc_path);
+		fclose(f);
 		return KNOT_ECRC;
 	}
 
-	unsigned long crc_from_file = 0;
-	if (fscanf(f_crc, "%lu\n", &crc_from_file) != 1) {
+	ssize_t crc_read_bytes = 0;
+	crc_read_bytes = read(f_crc, crc_buf, 64);
+	if (crc_read_bytes == -1) {
+		/* FIXME: Print strerror_r(errno) in dbg message */
 		dbg_zload("knot_zload_open: could not read "
 		                   "CRC from file '%s'\n",
 		                   crc_path);
-		fclose(f_crc);
+		free(crc_path);
+		close(f_crc);
+		fclose(f);
+		return KNOT_ECRC;
+	}
+	crc_buf[crc_read_bytes] = '\0';
+
+	unsigned long crc_from_file = 0;
+	errno = 0;
+	crc_from_file = strtoul(crc_buf, (char **) NULL, 10);
+	if (errno != 0) {
+		/* FIXME: Print strerror_r(errno) in dbg message */
+		dbg_zload("knot_zload_open: could not convert "
+		                   "CRC from file '%s'\n",
+		                   crc_path);
+		close(f_crc);
 		fclose(f);
 		free(crc_path);
 		return KNOT_ERROR;
 	}
+
+	/* Free some value and close the CRC file */
 	free(crc_path);
-	fclose(f_crc);
+	close(f_crc);
 
 	/* Compare calculated and read CRCs. */
 	if (crc_from_file != crc_calculated) {
@@ -894,12 +959,13 @@ static knot_dname_t **create_dname_array(FILE *f, uint max_id)
 
 	knot_dname_t **array =
 		malloc(sizeof(knot_dname_t *) * ( max_id + 1));
-	memset(array, 0, sizeof(knot_dname_t *) * (max_id + 1));
 	if (array == NULL) {
 		ERR_ALLOC_FAILED;
 		return NULL;
 	}
 
+	memset(array, 0, sizeof(knot_dname_t *) * (max_id + 1));	
+
 	for (uint i = 0; i < max_id - 1; i++) {
 		knot_dname_t *read_dname = read_dname_with_id(f);
 		if (read_dname == NULL) {
@@ -949,14 +1015,6 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 
 	knot_node_t *tmp_node;
 
-	/* Load the dname table. */
-	/* CLEANUP */
-//	const knot_dname_table_t *dname_table =
-//		create_dname_table(f, total_dnames);
-//	if (dname_table == NULL) {
-//		return NULL;
-//	}
-
 	uint32_t node_count;
 	uint32_t nsec3_node_count;
 	uint32_t auth_node_count;
@@ -998,6 +1056,7 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 	if (dname_table == NULL) {
 		ERR_ALLOC_FAILED;
 		cleanup_id_array(id_array, 1, total_dnames);
+		free(dname_table);
 		return NULL;
 	}
 
@@ -1008,6 +1067,7 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 			loader->filename);
 		cleanup_id_array(id_array, 1,
 				 node_count + nsec3_node_count + 1);
+		free(dname_table);
 		return NULL;
 	}
 
@@ -1018,6 +1078,8 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 		cleanup_id_array(id_array, 1,
 				 node_count + nsec3_node_count + 1);
 		dbg_zload("Failed to create new zone from apex!\n");
+		knot_node_free(&apex, 0);
+		free(dname_table);
 		return NULL;
 	}
 
@@ -1027,8 +1089,6 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 	/* Assign dname table to the new zone. */
 	contents->dname_table = dname_table;
 
-	/* CLEANUP */
-//	apex->prev = NULL;
 	knot_node_set_previous(apex, NULL);
 
 	knot_node_t *last_node = 0;
@@ -1050,8 +1110,6 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 			}
 
 			knot_node_set_previous(tmp_node, last_node);
-			/* CLEANUP */
-//			tmp_node->prev = last_node;
 
 			if (tmp_node->rrset_count &&
 			    (knot_node_is_deleg_point(tmp_node) ||
@@ -1065,8 +1123,7 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 		}
 	}
 
-	assert(knot_node_previous(knot_zone_contents_apex(contents), 0)
-	       == NULL);
+	assert(knot_node_previous(knot_zone_contents_apex(contents)) == NULL);
 
 	knot_node_set_previous(knot_zone_contents_get_apex(contents),
 	                         last_node);
@@ -1085,16 +1142,12 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 			fprintf(stderr, "!! cannot add first nsec3 node, "
 				"exiting.\n");
 			knot_zone_deep_free(&zone, 0);
-			free(id_array);
-			/* TODO this will leak dnames from id_array that were
-			 * not assigned. */
+			cleanup_id_array(id_array, node_count + 1,
+					 nsec3_node_count + 1);
 			return NULL;
 		}
 
 		knot_node_set_previous(nsec3_first, NULL);
-		/* CLEANUP */
-//		nsec3_first->prev = NULL;
-
 		last_node = nsec3_first;
 	}
 
@@ -1109,8 +1162,6 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 			}
 
 			knot_node_set_previous(tmp_node, last_node);
-			/* CLEANUP */
-//			tmp_node->prev = last_node;
 
 			last_node = tmp_node;
 		} else {
@@ -1120,10 +1171,8 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 	}
 
 	if (nsec3_node_count) {
-		assert(knot_node_previous(nsec3_first, 0) == NULL);
+		assert(knot_node_previous(nsec3_first) == NULL);
 		knot_node_set_previous(nsec3_first, last_node);
-		/* CLEANUP */
-//		nsec3_first->prev = last_node;
 	}
 
 	/* ID array is now useless */
@@ -1134,6 +1183,7 @@ knot_zone_t *knot_zload_load(zloader_t *loader)
 	free(id_array);
 
 	dbg_zload("zone loaded, returning: %p\n", zone);
+	
 	return zone;
 }
 
@@ -1142,18 +1192,19 @@ int knot_zload_needs_update(zloader_t *loader)
 	if (!loader) {
 		return 1;
 	}
+	
+        /* Check if the source still exists. */
+        struct stat st_src;
+        if (stat(loader->source, &st_src) != 0) {
+                return 1;
+        }
 
-	/* Check if the source still exists. */
-	struct stat st_src;
-	if (stat(loader->source, &st_src) != 0) {
-		return 1;
-	}
-
-	/* Check if the compiled file still exists. */
-	struct stat st_bin;
-	if (stat(loader->filename, &st_bin) != 0) {
-		return 1;
-	}
+        /* Check if the compiled file still exists. */
+        struct stat st_bin;
+        if (stat(loader->filename, &st_bin) != 0) {
+                return 1;
+        }
+	
 
 	/* Compare the mtime of the source and file. */
 	/*! \todo Inspect types on Linux. */
