@@ -81,9 +81,10 @@ typedef int (*xfr_callback_t)(int session, sockaddr_t *addr,
 typedef struct knot_ns_xfr {
 	int type;
 	int flags;
-	sockaddr_t addr;
+	sockaddr_t addr, saddr;
 	knot_packet_t *query;
 	knot_packet_t *response;
+	knot_rcode_t rcode;
 	xfr_callback_t send;
 	int session;
 	
@@ -100,7 +101,10 @@ typedef struct knot_ns_xfr {
 	size_t wire_size;
 	void *data;
 	knot_zone_t *zone;
+	char* zname;
 	void *owner;
+	knot_zone_contents_t *new_contents;
+	char *msgpref;
 	
 	/*! \note [TSIG] TSIG fields */
 	/*! \brief Message(s) to sign in wireformat. 
@@ -148,23 +152,21 @@ static const size_t KNOT_NS_TSIG_DATA_MAX_SIZE = 100 * 64 * 1024;
  */
 enum knot_ns_xfr_flag_t {
 	XFR_FLAG_TCP = 1 << 0, /*!< XFR request is on TCP. */
-	XFR_FLAG_UDP = 1 << 1  /*!< XFR request is on UDP. */
+	XFR_FLAG_UDP = 1 << 1,  /*!< XFR request is on UDP. */
+	XFR_FLAG_AXFR_FINISHED = 1 << 2
 };
 
 /*!
  * \brief XFR request types.
  */
 typedef enum knot_ns_xfr_type_t {
-	/* Special events. */
-	XFR_TYPE_CLOSE = -1, /*!< Close connection event. */
-
 	/* DNS events. */
-	XFR_TYPE_AIN = 0,   /*!< AXFR-IN request (start transfer). */
-	XFR_TYPE_AOUT,  /*!< AXFR-OUT request (incoming transfer). */
-	XFR_TYPE_IIN,   /*!< IXFR-IN request (start transfer). */
-	XFR_TYPE_IOUT,  /*!< IXFR-OUT request (incoming transfer). */
-	XFR_TYPE_SOA,   /*!< Pending SOA request. */
-	XFR_TYPE_NOTIFY /*!< Pending NOTIFY query. */
+	XFR_TYPE_AIN = 1 << 0,   /*!< AXFR-IN request (start transfer). */
+	XFR_TYPE_AOUT= 1 << 1,  /*!< AXFR-OUT request (incoming transfer). */
+	XFR_TYPE_IIN = 1 << 2,   /*!< IXFR-IN request (start transfer). */
+	XFR_TYPE_IOUT = 1 << 3,  /*!< IXFR-OUT request (incoming transfer). */
+	XFR_TYPE_SOA = 1 << 4,   /*!< Pending SOA request. */
+	XFR_TYPE_NOTIFY = 1 << 5 /*!< Pending NOTIFY query. */
 } knot_ns_xfr_type_t;
 
 /*----------------------------------------------------------------------------*/
@@ -175,6 +177,10 @@ typedef enum knot_ns_xfr_type_t {
  */
 knot_nameserver_t *knot_ns_create();
 
+/*! \todo Document me. */
+void knot_ns_set_nsid(knot_nameserver_t *nameserver, const char *nsid,
+                      size_t nsid_length);
+
 /*!
  * \brief Parses the given query into the response structure and recognizes
  *        type of the query.
