@@ -146,6 +146,47 @@ static const knot_dname_t *knot_rdata_srv_name(const knot_rdata_t *rdata)
 	return rdata->items[3].dname;
 }
 
+/*----------------------------------------------------------------------------*/
+
+static void knot_rdata_free_items(knot_rdata_item_t *items, unsigned int count,
+                                  uint type, int free_all_dnames)
+{
+	if (items == NULL) {
+		return;
+	}
+
+	knot_rrtype_descriptor_t *desc = knot_rrtype_descriptor_by_type(type);
+	assert(desc != NULL);
+
+	assert(count <= desc->length);
+
+	for (int i = 0; i < count; i++) {
+		if (&(items[i]) == NULL) {
+			continue;
+		}
+		if (desc->wireformat[i] == KNOT_RDATA_WF_COMPRESSED_DNAME
+		    || desc->wireformat[i] == KNOT_RDATA_WF_UNCOMPRESSED_DNAME
+		    || desc->wireformat[i] == KNOT_RDATA_WF_LITERAL_DNAME) {
+			if ((items[i].dname != NULL)) {
+				/*! \todo This is hack to prevent memory errors,
+				 *        as the rdata_set_items() cannot determine
+				 *        items type and so cannot increment
+				 *        reference count in case of dname type.
+				 *        Free would then release dnames that
+				 *        aren't referenced by the rdata.
+				 */
+				if (free_all_dnames) {
+					knot_dname_release(items[i].dname);
+				}
+			}
+		} else {
+			free(items[i].raw_data);
+		}
+	}
+
+	free(items);
+}
+
 /*----------------------------------------------------------------------------*/
 /* API functions                                                              */
 /*----------------------------------------------------------------------------*/
@@ -268,20 +309,41 @@ int knot_rdata_from_wire(knot_rdata_t *rdata, const uint8_t *wire,
 				break;
 			case 3:
 				pos2 = *pos;
-				fprintf(stderr, "reading dname from pos: %zu\n", pos2);
-				dname =
-					knot_dname_parse_from_wire(
+				//fprintf(stderr, "reading dname from pos: %zu\n", pos2);
+				dname = knot_dname_parse_from_wire(
 					         wire, &pos2, total_size, NULL);
 				if (dname == NULL) {
+					knot_rdata_free_items(items, i,
+					                      desc->type, 1);
 					return KNOT_ERROR;
 				}
-				items[i].dname = dname;
+			
+				items[i].raw_data = (uint16_t *)malloc(
+						knot_dname_size(dname) + 2);
+				if (items[i].raw_data == NULL) {
+					knot_dname_free(&dname);
+					knot_rdata_free_items(items, i,
+					                      desc->type, 1);
+					return KNOT_ENOMEM;
+				}
+				
+				items[i].raw_data[0] = knot_dname_size(dname);
+				memcpy((uint8_t *)(items[i].raw_data + 1),
+				       knot_dname_name(dname),
+				       knot_dname_size(dname));
+				
+//				items[i].dname = dname;
 				//*pos += dname->size;
 				parsed += pos2 - *pos;
 				
-				fprintf(stderr, "read %zu bytes.\n", parsed);
+				//fprintf(stderr, "read %zu bytes.\n", parsed);
 				*pos = pos2;
-				dname = 0;
+				knot_dname_free(&dname);
+				
+				// continue with next item
+				++i;
+				continue;
+				
 				break;
 			default:
 				assert(0);
@@ -289,41 +351,46 @@ int knot_rdata_from_wire(knot_rdata_t *rdata, const uint8_t *wire,
 
 			break;
 		default:
+			knot_rdata_free_items(items, i,
+			                      desc->type, 1);
 			return KNOT_EMALF;
 
 		}
 
 		if (item_size != 0) {
 			if (parsed + item_size > rdlength) {
-				free(items);
+				knot_rdata_free_items(items, i,
+				                      desc->type, 1);
 				return KNOT_EFEWDATA;
 			}
 
 			items[i].raw_data = (uint16_t *)malloc(item_size + 2);
 			if (items[i].raw_data == NULL) {
-				free(items);
+				knot_rdata_free_items(items, i,
+				                      desc->type, 1);
 				return KNOT_ENOMEM;
 			}
 			memcpy(items[i].raw_data, &item_size, 2);
 			memcpy(items[i].raw_data + 1, wire + *pos, item_size);
 			*pos += item_size;
 			parsed += item_size;
-		} else if (item_type == KNOT_RDATA_WF_BINARY
-		           || item_type == KNOT_RDATA_WF_IPSECGATEWAY) {
-			fprintf(stderr, "item_size was 0, creating empty rdata item.\n");
+		} else if (item_type == KNOT_RDATA_WF_BINARY/*
+		           || item_type == KNOT_RDATA_WF_IPSECGATEWAY*/) {
+//			fprintf(stderr, "item_size was 0, creating empty rdata item.\n");
 			// in this case we are at the end of the RDATA
 			// and should create an empty RDATA item
 			items[i].raw_data = (uint16_t *)malloc(2);
 			if (items[i].raw_data == NULL) {
-				free(items);
+				knot_rdata_free_items(items, i,
+				                      desc->type, 1);
 				return KNOT_ENOMEM;
 			}
 			memcpy(items[i].raw_data, &item_size, 2);
 		} else if (item_type != KNOT_RDATA_WF_COMPRESSED_DNAME
 		           && item_type != KNOT_RDATA_WF_UNCOMPRESSED_DNAME
 		           && item_type != KNOT_RDATA_WF_LITERAL_DNAME) {
-				fprintf(stderr, "RDATA item not set (i: %d), type: %u"
-					" RDATA item type: %d\n", i, desc->type ,item_type);
+//				fprintf(stderr, "RDATA item not set (i: %d), type: %u"
+//					" RDATA item type: %d\n", i, desc->type ,item_type);
 				assert(0);
 		}
 
@@ -474,39 +541,9 @@ void knot_rdata_deep_free(knot_rdata_t **rdata, uint type,
 		return;
 	}
 
-	knot_rrtype_descriptor_t *desc =
-		knot_rrtype_descriptor_by_type(type);
-	assert(desc != NULL);
+	knot_rdata_free_items((*rdata)->items, (*rdata)->count, type,
+	                      free_all_dnames);
 
-	assert((*rdata)->count <= desc->length);
-
-	for (int i = 0; i < (*rdata)->count; i++) {
-		if (&((*rdata)->items[i]) == NULL) {
-			continue;
-		}
-		if (desc->wireformat[i] == KNOT_RDATA_WF_COMPRESSED_DNAME
-		    || desc->wireformat[i] == KNOT_RDATA_WF_UNCOMPRESSED_DNAME
-		    || desc->wireformat[i] == KNOT_RDATA_WF_LITERAL_DNAME ) {
-			if (((*rdata)->items[i].dname != NULL)) {
-				/*! \todo This is hack to prevent memory errors,
-				 *        as the rdata_set_items() cannot determine
-				 *        items type and so cannot increment
-				 *        reference count in case of dname type.
-				 *        Free would then release dnames that
-				 *        aren't referenced by the rdata.
-				 */
-				if (free_all_dnames) {
-					knot_dname_release((*rdata)->items[i].dname);
-				}
-			}
-		} else {
-			free((*rdata)->items[i].raw_data);
-		}
-	}
-
-	if ((*rdata)->items) {
-		free((*rdata)->items);
-	}
 	free(*rdata);
 	*rdata = NULL;
 }
@@ -634,7 +671,7 @@ void knot_rdata_deep_free(knot_rdata_t **rdata, uint type,
 /*----------------------------------------------------------------------------*/
 
 knot_rdata_t *knot_rdata_deep_copy(const knot_rdata_t *rdata, 
-                                       uint16_t type)
+                                       uint16_t type, int copy_dnames)
 {
 	knot_rdata_t *copy = knot_rdata_new();
 	CHECK_ALLOC_LOG(copy, NULL);
@@ -658,8 +695,13 @@ knot_rdata_t *knot_rdata_deep_copy(const knot_rdata_t *rdata,
 		if (d->wireformat[i] == KNOT_RDATA_WF_COMPRESSED_DNAME
 		    || d->wireformat[i] == KNOT_RDATA_WF_UNCOMPRESSED_DNAME
 		    || d->wireformat[i] == KNOT_RDATA_WF_LITERAL_DNAME) {
-			copy->items[i].dname =
-				knot_dname_deep_copy(rdata->items[i].dname);
+			if (copy_dnames) {
+				copy->items[i].dname =
+				    knot_dname_deep_copy(rdata->items[i].dname);
+			} else {
+				copy->items[i].dname = rdata->items[i].dname;
+				knot_dname_retain(rdata->items[i].dname);
+			}
 		} else {
 			copy->items[i].raw_data = (uint16_t *)malloc(
 					rdata->items[i].raw_data[0] + 2);
