@@ -49,15 +49,17 @@ typedef struct knot_zone_contents_t {
 	knot_zone_tree_t *nodes;
 	knot_zone_tree_t *nsec3_nodes;
 
-	/*!
+	knot_dname_table_t *dname_table;
+
+	struct knot_zone *zone;
+
+	knot_nsec3_params_t nsec3_params;
+	
+  /*!
 	 * \todo Unify the use of this field - authoritative nodes vs. all.
 	 */
 	uint node_count;
 
-	knot_dname_table_t *dname_table;
-
-	knot_nsec3_params_t nsec3_params;
-
 	/*! \brief Generation of the zone during update.
 	 * 
 	 * Possible values:
@@ -67,8 +69,6 @@ typedef struct knot_zone_contents_t {
 	 *        used, no matter their generation.
 	 */
 	short generation;
-
-	struct knot_zone *zone;
 } knot_zone_contents_t;
 
 /*----------------------------------------------------------------------------*/
@@ -338,8 +338,7 @@ int knot_zone_contents_find_nsec3_for_name(
                                     const knot_zone_contents_t *contents,
                                     const knot_dname_t *name,
                                     const knot_node_t **nsec3_node,
-                                    const knot_node_t **nsec3_previous,
-				    int check_ver);
+                                    const knot_node_t **nsec3_previous);
 /*!
  * \brief Returns the apex node of the zone.
  *
@@ -362,7 +361,9 @@ knot_node_t *knot_zone_contents_get_apex(
  *
  * \param zone Zone to adjust domain names in.
  */
-int knot_zone_contents_adjust(knot_zone_contents_t *contents, int check_ver);
+int knot_zone_contents_adjust(knot_zone_contents_t *contents);
+
+int knot_zone_contents_check_loops(knot_zone_contents_t *zone);
 
 /*!
  * \brief Parses the NSEC3PARAM record stored in the zone.
@@ -544,13 +545,24 @@ int knot_zone_contents_dname_table_apply(knot_zone_contents_t *contents,
  * \retval KNOT_ENOMEM
  */
 int knot_zone_contents_shallow_copy(const knot_zone_contents_t *from,
-                                      knot_zone_contents_t **to);
+                                    knot_zone_contents_t **to);
+
+int knot_zone_contents_shallow_copy2(const knot_zone_contents_t *from,
+                                     knot_zone_contents_t **to);
+
+//int knot_zone_contents_dnames_from_node_to_table(
+//	knot_dname_table_t *table, knot_node_t *node);
+
+//void knot_zone_contents_adjust_node(knot_node_t *node,
+//                                    knot_zone_contents_t *zone, int check_ver);
 
 void knot_zone_contents_free(knot_zone_contents_t **contents);
 
 void knot_zone_contents_deep_free(knot_zone_contents_t **contents,
                                   int destroy_dname_table);
 
+int knot_zone_contents_integrity_check(const knot_zone_contents_t *contents);
+
 #endif
 
 /*! @} */
