@@ -23,6 +23,7 @@
 #include "libknot/util/error.h"
 #include "libknot/util/wire.h"
 #include "libknot/tsig-op.h"
+#include "common/print.h"
 
 #include "tsig_tests.h"
 
@@ -85,7 +86,7 @@ static int test_knot_tsig_sign()
 	int errors = 0;
 	/* Test bad arguments. */
 	int lived = 0;
-	lives_ok_silent({
+	lives_ok({
 		int ret = knot_tsig_sign(NULL, NULL, 0, NULL, 0, NULL,
 		               NULL, NULL, 0, 0);
 		if (ret != KNOT_EBADARG) {
@@ -142,7 +143,7 @@ static int test_knot_tsig_sign()
 			errors++;
 		}
 		lived = 1;
-	}, NULL);
+	}, "");
 
 	errors += !lived;
 	
@@ -239,8 +240,8 @@ static int test_knot_tsig_sign()
 	knot_wire_set_arcount(msg, 0);
 	/* Wire now should be identical. Compare with its pre-signing copy. */
 	if (strncmp((char *)msg, (char *)msg_copy, msg_len) != 0) {
-		hex_print(msg, msg_len);
-		hex_print(msg_copy, msg_len);
+		hex_print((const char*)msg, msg_len);
+		hex_print((const char*)msg_copy, msg_len);
 		diag("knot_tsig_sign has changed the signed wire!");
 		errors++;
 	}
@@ -267,8 +268,8 @@ static int test_knot_tsig_sign()
 	knot_wire_set_arcount(msg, 0);
 	/* Wire now should be identical. Compare with its pre-signing copy. */
 	if (strncmp((char *)msg, (char *)msg_copy, msg_len) != 0) {
-		hex_print(msg, msg_len);
-		hex_print(msg_copy, msg_len);
+		hex_print((const char*)msg, msg_len);
+		hex_print((const char*)msg_copy, msg_len);
 		diag("knot_tsig_sign has changed the signed wire!");
 		errors++;
 	}
@@ -300,7 +301,7 @@ static int test_knot_tsig_sign_next()
 	int errors = 0;
 	/* Test bad arguments. */
 	int lived = 0;
-	lives_ok_silent({
+	lives_ok({
 		int ret = knot_tsig_sign_next(NULL, NULL, 0, NULL, 0, NULL,
 		               NULL, NULL, NULL, 0); /*! \todo FIX */
 		if (ret != KNOT_EBADARG) {
@@ -358,7 +359,7 @@ static int test_knot_tsig_sign_next()
 			errors++;
 		}
 		lived = 1;
-	}, NULL);
+	}, "");
 
 	errors += !lived;
 	
@@ -415,7 +416,7 @@ static int test_knot_tsig_server_check()
 	int errors = 0;
 	/* Test bad arguments. */
 	int lived = 0;
-	lives_ok_silent({
+	lives_ok({
 		int ret = knot_tsig_server_check(NULL, NULL, 0, NULL);
 		if (ret != KNOT_EBADARG) {
 			diag("NULL argument did not return KNOT_EBADARG!");
@@ -432,7 +433,7 @@ static int test_knot_tsig_server_check()
 			errors++;
 		}
 		lived = 1;
-	}, NULL);
+	}, "");
 		
 	errors += !lived;
 	
@@ -523,7 +524,7 @@ static int test_knot_tsig_client_check()
 			errors++;
 		}
 		lived = 1;
-	}, NULL);
+	}, "");
 		
 	errors += !lived;
 	
@@ -631,7 +632,7 @@ static int test_knot_tsig_test_tsig_add()
 			errors++;
 		}
 		lived = 1;
-	}, NULL);
+	}, "");
 	
 	errors += !lived;
 	
