@@ -1463,14 +1463,16 @@ uint16_t * zparser_conv_services(const char *protostr, char *servicestr)
 
 	char *sp = 0;
 	while ((word = strtok_r(servicestr, sep, &sp))) {
-		struct servent *service;
+		struct servent *service = NULL;
 		int port;
 
 		service = getservbyname(word, proto->p_name);
 		if (service) {
 			/* Note: ntohs not ntohl!  Strange but true.  */
 			port = ntohs((uint16_t) service->s_port);
+//			printf("assigned port %d\n", port);
 		} else {
+//			printf("else\n");
 			char *end;
 			port = strtol(word, &end, 10);
 			if (*end != '\0') {
@@ -1479,7 +1481,7 @@ uint16_t * zparser_conv_services(const char *protostr, char *servicestr)
 					" protocol '%s'",
 					word, protostr);
 				parser->error_occurred = KNOTDZCOMPILE_EBRDATA;
-				continue;
+				return NULL;
 			}
 		}
 
@@ -1491,6 +1493,7 @@ uint16_t * zparser_conv_services(const char *protostr, char *servicestr)
 				max_port = port;
 			}
 		}
+		servicestr = NULL;
 	}
 
 	r = alloc_rdata(sizeof(uint8_t) + max_port / 8 + 1);
@@ -1502,7 +1505,7 @@ uint16_t * zparser_conv_services(const char *protostr, char *servicestr)
 
 	p = (uint8_t *)(r + 1);
 	*p = proto->p_proto;
-	memcpy(p + 1, bitmap, *r);
+	memcpy(p + 1, bitmap, *r - 1);
 
 	return r;
 }
@@ -2060,7 +2063,7 @@ uint16_t * zparser_conv_loc(char *str)
 				return NULL;
 			}
 
-			if (sscanf(start, "%lf", &d) != 1) {
+			if (sscanf(start, "%16lf", &d) != 1) {
 				zc_error_prev_line("error parsing seconds");
 			}
 
@@ -2158,7 +2161,7 @@ uint16_t * zparser_conv_loc(char *str)
 		++str;
 	}
 
-	if (sscanf(start, "%lf", &d) != 1) {
+	if (sscanf(start, "%16lf", &d) != 1) {
 		zc_error_prev_line("error parsing altitude");
 	}
 
@@ -2334,6 +2337,10 @@ void zadd_rdata_wireformat(uint16_t *data)
  */
 void zadd_rdata_txt_wireformat(uint16_t *data, int first)
 {
+	if (data == NULL) {
+		parser->error_occurred = KNOTDZCOMPILE_EBRDATA;
+		return;
+	}
 	dbg_zp("Adding text!\n");
 //	hex_print(data + 1, data[0]);
 	knot_rdata_item_t *rd;
@@ -2358,6 +2365,10 @@ void zadd_rdata_txt_wireformat(uint16_t *data, int first)
 		rd = &parser->temporary_items[parser->rdata_count-1];
 	}
 
+	if (rd == NULL || rd->raw_data == NULL) {
+		return;
+	}
+
 	if ((size_t)rd->raw_data[0] + (size_t)data[0] > 65535) {
 		zc_error_prev_line("too large rdata element");
 		return;
